import logging
import os.path
import os.path
import typing
from typing import Union, Callable

import yaml
from PySide6.QtGui import Qt, QColor, QBrush
from PySide6.QtWidgets import QDialog, QVBoxLayout, QPushButton, QLabel, QListWidget, QLineEdit, QListWidgetItem, \
    QHBoxLayout, QWidget, QTabWidget, QTableWidget, QTableWidgetItem, \
    QHeaderView, QColorDialog
from sqlalchemy.util import await_only

from famegui.config.config import BASE_DIR
from famegui.config.style_config import LIST_ITEM_STYLE
from famegui.database.db_loader.db_manager import get_existing_table_names, read_all_data, update_db, insert
from famegui.database.settings_db.init_db import Setting, init_db_session
from famegui.famegui_widgets.StringSetSelectWidget import StringSetSelectWidget
from famegui.models import Scenario


class EditFameDbValuesDialog(QDialog):
    settings_data = {}
    agent_type_to_table_cell: typing.Dict[str, dict[str:typing.Union[QTableWidgetItem, str]]] = {}

    def save_settings_to_db(self):

        with init_db_session() as session:
            for setting_name, value in self.settings_data.items():
                setting = session.query(Setting).filter_by(setting_name=setting_name).first()
                setting.value = value

            session.commit()

    def load_dialog(self):
        with init_db_session() as session:
            yaml_file = os.path.join(BASE_DIR, "config", 'settings_config.yaml')  # Path to your YAML file
            with open(yaml_file, 'r') as file:
                data = yaml.safe_load(file)
            # Extract and insert data

            for table_name, table_data in data['db']['tables'].items():
                if table_name == 'setting':
                    table_fields = data['db']['tables']['setting']["preset_values"]

                    for item in table_fields:
                        setting = session.query(Setting).filter_by(
                            setting_name=item['setting_name']).first()

                        if not setting:
                            new_setting = Setting(
                                value=item['value'],
                                setting_name=item['setting_name'],
                                setting_label=item['setting_label'],
                                data_type=item['data_type'])
                            session.add(new_setting)
                        session.commit()

    def load_settings(self):

        with init_db_session() as session:
            settings = session.query(Setting).all()
            for setting in settings:
                # Create the custom widget
                widget = QWidget()
                layout = QHBoxLayout(widget)
                widget.setStyleSheet(LIST_ITEM_STYLE)
                layout.setContentsMargins(0, 0, 0, 20)  # Remove margins

                label = QLabel(setting.setting_label)
                label.setWordWrap(True)  # Enable word wrap for the label

                input = QLineEdit()

                input.textChanged.connect(lambda text: self.on_text_changed(text, setting.setting_name))

                input.setText(setting.value)
                input.setFixedWidth(200)
                input.setStyleSheet("border: 1px solid lightgray; padding: 5px; border-radius: 4px;")

                layout.addWidget(label)
                layout.addWidget(input)

                # Create a QListWidgetItem
                item = QListWidgetItem(self.list_widget)

                # Set the custom widget as the item widget
                self.list_widget.setItemWidget(item, widget)

                # Store the setting object in the item's data
                item.setData(Qt.UserRole, setting)

                # Set a default size for the item
                item.setSizeHint(widget.sizeHint())

    def on_string_sets_saved(self, data_dict: dict):
        with init_db_session() as session:
            for string_set_group_name, string_set_list in data_dict.items():
                for string_set_value in string_set_list:
                    insert(session, "db_string_item", {
                        "parent_string_set_group_name": string_set_group_name,
                        "string_set_item_value": string_set_value
                    })

    def get_strings_set_tab(self) -> QWidget:
        string_set_tab = QWidget()
        string_set_layout = QVBoxLayout()
        string_set_main_container = QWidget()
        string_set_main_container.setStyleSheet("background-color:#A7BBEC")
        string_set_widget = StringSetSelectWidget(self._scenario)
        string_set_widget.on_string_sets_saved.connect(self.on_string_sets_saved)

        string_set_layout.addWidget(string_set_widget, stretch=1)
        string_set_tab.setLayout(string_set_layout)

        return string_set_tab

    def __init__(self, scenario: Scenario,
                 handle_agent_color_change: Callable):
        super().__init__()
        self._scenario = scenario

        self._scenario.string_sets.update()

        self._handle_agent_color_change = handle_agent_color_change

        self.setStyleSheet("""
               QDialog {
                   background-color: #f5f5f5;
               }
               QListWidget {
                   border: 1px solid lightgray;
                   border-radius: 4px;
               }
               QLabel {
                   font-size: 14px;
               }
           """)

        self.setMinimumHeight(400)
        self.setMinimumWidth(600)

        self.setWindowTitle("Settings")
        self.layout = QVBoxLayout()

        self.list_widget = QListWidget()
        self.list_widget.currentItemChanged.connect(self.display_setting)
        self.layout.addWidget(self.list_widget)

        self.input_layout = QVBoxLayout()
        self.layout.addLayout(self.input_layout)
        self.input_layout.setContentsMargins(20, 20, 20, 20)
        self.input_layout.setSpacing(15)

        self.load_dialog()
        self.save_button = QPushButton("Save")
        self.save_button.clicked.connect(self.save_setting)
        self.layout.addWidget(self.save_button)
        self.row_idx_to_db_id_mapper = {}

        tab_holder = QVBoxLayout()

        tabs = QTabWidget()

        # Create tab 1
        tab1 = QWidget()
        tab1_layout = QVBoxLayout()
        tab1_layout.addWidget(self.list_widget)
        tab1.setLayout(tab1_layout)

        # Create tab 2
        yaml_file = os.path.join(BASE_DIR, "config", 'settings_config.yaml')  # Path to your YAML file

        with init_db_session() as session:
            table_name_list = get_existing_table_names(yaml_file, session)

            table_name_list = [item for item in table_name_list if
                               item != "db_string_set_group" and item != "db_string_item"]

            data = read_all_data(session)

            string_set_tab = self.get_strings_set_tab()

            tabs.addTab(string_set_tab, "String Set")

            for table_name in table_name_list:
                table_specific_tab = QWidget()
                table_specific_tab_layout = QVBoxLayout()
                table_widget = self.create_table(data[table_name], table_name)

                table_specific_tab_layout.addWidget(table_widget)

                table_specific_tab.setLayout(table_specific_tab_layout)
                tabs.addTab(table_specific_tab, table_name)

            tab_holder.addWidget(tabs)

            self.setLayout(tab_holder)

            self.load_settings()

    def reject(self):
        """Acts like on close due too no explicit closing/accepting button"""

        color_settings_data = {setting_item_key: setting_item_value for
                               setting_item_key, setting_item_value in
                               self.settings_data.items() if setting_item_key.__contains__("hex")}

        if len(color_settings_data.keys()) != 0:
            self._handle_agent_color_change()

        self.settings_data.clear()
        self.agent_type_to_table_cell.clear()

        super().reject()

    def save_setting(self):

        self.save_settings_to_db()

        # quit dlg

        self.close()

    def on_table_item_changed(self, item: Union[QTableWidgetItem, dict], table_name: str):
        """Handles changes to table items and updates settings_data."""

        row, column, new_value = None, None, None

        db_field_name = None

        if isinstance(item, QTableWidgetItem):
            row, column, new_value = item.row(), item.column(), item.text()

        if isinstance(item, dict):
            row, column, new_value, db_field_name = item["row"], item["column"], item["text"], item["db_field_name"]

        # Get the setting name (column header) and update settings_data
        field_name = item.tableWidget().horizontalHeaderItem(column).text() if db_field_name is None else db_field_name

        # Assuming your data is a list of dictionaries with row-wise settings
        self.settings_data[field_name] = new_value

        updated_record = None

        with init_db_session() as session:
            db_id = self.row_idx_to_db_id_mapper[table_name][row]

            updated_record = update_db(session, table_name, db_id, field_name,
                                       new_value, return_updated_record=True)

        logging.info(f"Updated setting {field_name} at row {row} to {new_value}")

        if not field_name.__contains__("hex"):
            return

        agent_type_name = updated_record["agent_type_name"]

        self.agent_type_to_table_cell[agent_type_name]["cell_widget"].setBackground(QBrush(new_value))

    def open_color_picker(self, row, col, local_value, color_line_edit: QLineEdit):

        color_dlg = QColorDialog(QColor(color_line_edit.text()))
        if color_dlg.exec() != 0:
            selected_color = color_dlg.selectedColor()

            if selected_color is not None:
                hex_without_hash = selected_color.name()[1:]

                hex_without_hash = "#" + hex_without_hash
                color_line_edit.setText(hex_without_hash)

    def get_color_edit_panel(self, initial_color: str, row, cell_idx, table_name, db_field_name) -> [QWidget]:
        container = QWidget()
        container.setFixedHeight(50)
        h_layout = QHBoxLayout()
        container.setLayout(h_layout)
        color_line_edit = QLineEdit(initial_color)
        h_layout.addWidget(color_line_edit)

        btn = QPushButton(f"Pick color : {table_name}")
        h_layout.addWidget(btn)

        color_line_edit.textChanged.connect(lambda new_str: self.on_table_item_changed({
            "row": row,
            "column": cell_idx,
            "text": new_str,
            "db_field_name": db_field_name

        }, table_name))

        btn.clicked.connect(lambda x: self.open_color_picker(row, cell_idx, str(initial_color), color_line_edit))

        return container

    def create_table(self, table_data: [], table_name):
        """Creates a table widget with editable fields for the settings."""
        table = QTableWidget()

        # Two columns: Setting Name and Value

        if len(table_data) == 0:
            return table

        field_names = [field_key for field_key in table_data[0].keys()]

        row_idx_to_ids = {idx: item["id"] for idx, item in enumerate(table_data)}

        self.row_idx_to_db_id_mapper[table_name] = row_idx_to_ids

        field_names = [item for item in field_names if
                       not item.__contains__("id") and not item.__contains__("data_type")]

        table.setColumnCount(len(field_names))
        table.setRowCount(len(table_data))

        table.setHorizontalHeaderLabels(field_names)

        for row, setting in enumerate(table_data):
            # Create the setting name item

            row_values = [setting[item] for item in field_names]

            for idx, item in enumerate(row_values):

                if field_names[idx].lower() == "agent_type_name":
                    painted_cell = QTableWidgetItem(str(item))
                    color_value = row_values[idx - 1]
                    background_color = QColor(color_value)

                    self.agent_type_to_table_cell[str(item)] = {
                        "cell_widget": painted_cell,
                        "agent_type_name": str(item)
                    }

                    brush = QBrush(background_color)
                    painted_cell.setBackground(brush)

                    table.setItem(row, idx, painted_cell)
                    continue

                if field_names[idx].lower().__contains__("hex"):
                    table.setRowHeight(row, 60)

                    db_field_name = field_names[idx]

                    color_input_panel = self.get_color_edit_panel(str(item), row, idx,
                                                                  table_name, db_field_name)
                    table.setItem(row, idx, QTableWidgetItem())

                    table.setCellWidget(row, idx, color_input_panel)
                    continue

                table.setItem(row, idx, QTableWidgetItem(str(item)))

            # Connect the value item change to update the settings_data

        table.itemChanged.connect(lambda item: self.on_table_item_changed(item, table_name))

        # Resize columns to fit content
        table.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.Stretch)
        return table

    def display_setting(self, current, previous):
        for i in reversed(range(self.input_layout.count())):
            widget = self.input_layout.itemAt(i).widget()
            if widget is not None:
                widget.setParent(None)

    def on_text_changed(self, text, setting_name):
        self.settings_data[setting_name] = text

    def on_button_click(self):
        self.label.setText("Button Clicked!")



import logging
import os

from PySide6 import QtWidgets, QtCore, QtUiTools
from PySide6.QtCore import QFile
from PySide6.QtWidgets import QLabel, QLineEdit, QHBoxLayout, QPushButton, QComboBox, QDateTimeEdit

from famegui.config.contract_consts import PRESET_DURATION_IN_SECONDS
from famegui.config.runtime_consts import UPDATED_CONTRACT_KEY, OLD_CONTRACT_KEY, \
    CONTRACT_FIELD_DELIVERY_INTERVAL_IN_STEPS
from famegui.database.prebuild_queries.interval_presets import get_delivery_interval_presets
from famegui.maincontroller import MainController
from famegui.models import Contract, Scenario, Agent
from famegui.ui.fame_input_panels import (
    get_string_chooser_panel,
    FameCalendarInputPanelWidget,
)
from famegui.ui.fame_ui_elements import QFameBoldLabel
from famegui.ui.helper import gen_preset_interval_chooser
from famegui.ui.quick_modals import gen_quick_warning_modal
from famegui.utils import contract_fields_to_natural_types

DISPLAY_DATE_FORMAT = "yyyy-MM-dd"

DISPLAY_DATETIME_FORMAT = "yyyy-MM-dd HH:mm:ss"


class EditContractDialog(QtWidgets.QDialog):
    """Dialog for editing a contract"""

    on_contract_edited = QtCore.Signal(dict)

    def __init__(
            self,
            selected_contract: Contract,
            scenario: Scenario,
            controller: MainController,
            on_contract_edited_slot,
            parent=None,
    ):
        super(EditContractDialog, self).__init__(parent)

        # Set up the user interface from Designer
        self._selected_contract = selected_contract
        self._scenario = scenario
        self._controller = controller
        self.on_contract_edited.connect(on_contract_edited_slot)

        self._delivery_interval_presets = get_delivery_interval_presets()

        # Load UI
        loader = QtUiTools.QUiLoader()
        file = QFile(os.path.join(os.path.dirname(__file__), "edit_contract_dialog.ui"))
        file.open(QtCore.QFile.OpenModeFlag.ReadOnly)
        self.ui = loader.load(file, self)

        file.close()

        self._input_fields = []
        self.ui.verticalLayout: QtWidgets.QVBoxLayout

        self._init_static_ui()
        self._generate_input_items()
        self.create_btn_panel()

        self.ui.exec_()

    def setup_text_content(self):
        """Set up the headline content of the dialog"""
        title = QLabel(
            f"Edit Contract"
            f"\nProduct Name: {self._selected_contract.product_name}"
            f"\nSender Id: {self._selected_contract.sender_id}"
            f"\nReceiver Id: {self._selected_contract.receiver_id}",
            self,
        )

        title.setStyleSheet("font-size: 18px; font-weight: bold;")
        self.ui.verticalLayout.addWidget(title)

        description = QLabel("Leave blank if no value is required", self)
        description.setWordWrap(True)

        self.ui.verticalLayout.addWidget(description)

    def setup_layout(self):
        """Set up the layout frame of the dialog"""
        self.ui.verticalLayout.setContentsMargins(20, 20, 20, 20)
        self.ui.verticalLayout.setSpacing(20)

    def _init_static_ui(self):
        """Set up the static UI elements of the dialog"""
        self.setup_layout()
        self.setup_text_content()

    def _construct_product_chooser(self, layout, field_name):
        """construct product chooser widget"""

        sender_type = self._scenario.schema.agent_type_from_name(
            self._controller.get_agent_ctrl(self._selected_contract.sender_id).type_name
        )

        sender_type_products = sender_type.products

        selected_contract = self._selected_contract

        selected_contract_dict = selected_contract.to_dict()

        current_product_name = selected_contract_dict[
            "productname"] if "productname" in selected_contract_dict else None

        inner_layout, comboBox = get_string_chooser_panel(sender_type_products, Contract.KEY_PRODUCT,
                                                          preset_value=current_product_name)

        layout.addLayout(inner_layout)
        self._input_fields.append((field_name, comboBox))

    def _construct_date_time_chooser(self, layout, field_name, contract_data):
        """construct product time chooser widget"""

        saved_fame_datetime = None
        if field_name in contract_data:
            saved_fame_datetime = contract_data[field_name]

        fame_calender_input_panel = FameCalendarInputPanelWidget(
            self, field_name, saved_fame_datetime
        )
        (
            inner_layout,
            date_time_edit,
        ) = fame_calender_input_panel.get_input_related_widgets()
        layout.addWidget(fame_calender_input_panel)

        self._input_fields.append((field_name, date_time_edit))

    def on_preset_changed(self, index, selected_option: str):
        """
        Handles the change event when a new preset is selected
        """
        # Do something with the selected preset

        for field_name, input in self._input_fields:

            if field_name == CONTRACT_FIELD_DELIVERY_INTERVAL_IN_STEPS:

                if selected_option == "Custom":
                    continue
                input.setText(self._delivery_interval_presets[index][PRESET_DURATION_IN_SECONDS])

    def _construct_default_input_field(self, layout, field_name, contract_data):
        """construct default input field"""
        label = QFameBoldLabel(text=field_name)
        label.setMinimumWidth(int(self.width() * 0.3))  # Set the maximum width to 30% of the dialog's width
        label.setMaximumWidth(int(self.width() * 0.3))  # Set the maximum width to 30% of the dialog's width
        label.setWordWrap(True)  # Enable word wrap

        input_field = QLineEdit()

        preset_value = contract_data[field_name] if field_name in contract_data else ""

        input_field.setText(str(preset_value))

        inner_layout = QHBoxLayout()
        inner_layout.addWidget(label)

        choices = [item["preset_label"] for item in self._delivery_interval_presets]

        choices_bundles = [item for item in self._delivery_interval_presets]

        # self._selected_contract.delivery_interval

        if field_name.lower() == CONTRACT_FIELD_DELIVERY_INTERVAL_IN_STEPS:

            matched_choice = next(
                (item for item in self._delivery_interval_presets if
                 item["preset_duration_in_seconds"] == str(preset_value)),
                "Custom")

            matched_choice = matched_choice["preset_label"] if isinstance(matched_choice, dict) else matched_choice

            interval_preset_chooser = gen_preset_interval_chooser(choices, self.on_preset_changed,
                                                                  input_field, choices_bundles, matched_choice)

            inner_layout.addWidget(interval_preset_chooser)

        inner_layout.addWidget(input_field)

        layout.addLayout(inner_layout)
        self._input_fields.append((field_name, input_field))

    def _generate_input_items(self):
        """Entry point for generating input items"""

        layout = self.ui.verticalLayout
        contract_data = self._selected_contract.to_dict()

        # Create input rows dynamically based on the item list
        for field_name in self._selected_contract.get_fields_to_edit():
            if field_name == Contract.KEY_PRODUCT:
                self._construct_product_chooser(layout, field_name)

                continue

            field_type = self._selected_contract.get_field_type(field_name)

            if field_type == self._selected_contract.FIELD_TYPE_DATE_TIME:
                self._construct_date_time_chooser(layout, field_name, contract_data)

                continue
            self._construct_default_input_field(layout, field_name, contract_data)

    def create_btn_panel(self):
        """Construct the button panel"""
        cancel_button = QPushButton("Cancel", self)
        confirm_button = QPushButton("Confirm", self)
        cancel_button.clicked.connect(self.cancel)
        confirm_button.clicked.connect(self.confirm)

        # Create a QHBoxLayout for the buttons
        button_layout = QHBoxLayout()

        # Add the buttons to the button layout
        button_layout.addWidget(cancel_button)
        button_layout.addWidget(confirm_button)

        self.ui.verticalLayout.addLayout(button_layout)

    def init_ui(self):
        input = QLineEdit(self)
        self.ui.verticalLayout.addWidget(input)

    def cancel(self):
        """Cancel the applied changes and close the dialog"""
        self.ui.close()

    def _check_if_agent_exists(self, agent_id) -> bool:

        if not self._scenario.agent_exists(agent_id):
            gen_quick_warning_modal(
                self, "Agent does not exist", "Agent with id " + str(agent_id) + " does not exist"
            )
            return False

        return True

    def _check_for_product_support(self, agent: Agent) -> bool:

        if not self._scenario.schema.agent_supports_product(
                self._scenario.schema.agent_types[agent.type_name], self._selected_contract.product_name
        ):
            message = f"Agent with id '{str(agent.id)}' does not support the product: {self._selected_contract.product_name}"
            gen_quick_warning_modal(self, "Sender does not support product", message)
            return False
        return True

    def confirm(self):
        """Confirm the applied changes and close the dialog"""
        data_dict = {}

        for field_name, input_field in self._input_fields:
            if isinstance(input_field, QLineEdit):
                input_field: QLineEdit
                data_dict[field_name] = input_field.text()

                try:
                    input_data = int(input_field.text())
                except Exception as exce:
                    logging.debug(F"OPTIONAL FIELD NOT SET {str(exce)}")
                    continue

                if field_name == Contract.KEY_SENDER or field_name == Contract.KEY_SENDER:
                    if not self._scenario.agent_exists(input_data):
                        if not self._check_if_agent_exists(input_data):
                            return

                    agent: Agent = self._scenario.get_agent_by_id(input_data)

                    if not self._check_for_product_support(agent):
                        return

            if isinstance(input_field, QDateTimeEdit):
                input_field: QDateTimeEdit
                data_dict[field_name] = input_field.dateTime().toString(DISPLAY_DATETIME_FORMAT)
            if isinstance(input_field, QComboBox):
                input_field: QComboBox
                data_dict[field_name] = input_field.currentText()

        old_agent_dict = self._selected_contract.to_dict()

        dict_to_send = {
            OLD_CONTRACT_KEY: old_agent_dict.copy(),
        }

        for key, value in data_dict.items():
            old_agent_dict[key] = value
        old_agent_dict = contract_fields_to_natural_types(
            old_agent_dict, self._selected_contract
        )
        updated_contract = Contract.from_dict(old_agent_dict)

        dict_to_send[UPDATED_CONTRACT_KEY] = updated_contract.to_dict()

        self._scenario.update_contract(
            old_contract=self._selected_contract, updated_contract=updated_contract
        )
        self._controller.set_unsaved_changes(True)
        self.on_contract_edited.emit(dict_to_send)
        self.ui.accept()

    @staticmethod
    def select_date(date, line_edit):
        selected_date = date.toString(DISPLAY_DATE_FORMAT)
        line_edit.setText(selected_date)



import re
from datetime import datetime

from PySide6.QtWidgets import QWidget, QVBoxLayout, QLineEdit, QFrame
from fameio.input.schema import AttributeSpecs, AttributeType
from icecream import ic

from famegui.ui.block_item_wrapper import BlockItemWrapper


class BlockTypeWrapper(QFrame):
    """Wrapper for all types of Block input panels. It is used to validate input and style it accordingly to their type.
    This Wrapper is used for the BlockType input panels compared to the primitive input panels.
    """

    def __init__(
            self,
            parent=None,
            block_type_layout=None,
            spec: AttributeSpecs = None,
            attr_name=None,
            dict_data=None,
    ):
        super().__init__(parent)
        self.block_type_layout: QVBoxLayout = block_type_layout
        self.setLayout(self.block_type_layout)

        matches = re.findall(r"(?<=\.)[^.]*$", spec.full_name)

        attr_short_name = matches[0]
        self.attr_name = attr_short_name

        self.dict_data = dict_data

        self.attr_specs: AttributeSpecs = spec
        self.parent = parent
        uniq_obj_name = f"myUniqueWidget-{id(self)}"
        self.setObjectName(uniq_obj_name)
        color_str = "rgba(255, 221, 148, 0.4)"

        stylesheet = (
            f"#{uniq_obj_name} {{ background-color: {color_str};  border-radius: 5px;}}"
        )

        self.setStyleSheet(stylesheet)

    def get_attr_full_name(self):
        return self.attr_specs.full_name

    def get_attr_name(self):
        """Get the attribute short name."""
        return self.attr_name

    def _process_block_type_input(self, widget, data_list, data):
        widget: BlockTypeWrapper
        inner_data = {}
        for key, value in widget.get_data().items():
            inner_data[key] = value
        data_list.append(inner_data)

        if widget.attr_name in widget.get_data():
            inner_data = {}
            for key, value in widget.get_data().items():
                inner_data[key] = value

        if self.get_attr_name() in widget.get_data():
            for key, value in widget.get_data().items():
                data[key] = value
            return

        data[widget.get_attr_name()] = inner_data

    def _process_block_item_type_input(self, widget, data_list, data):
        widget: BlockItemWrapper

        inner_data = {}
        for key, value in widget.get_data().items():
            inner_data[key] = value

        data[widget.get_short_attr_name()] = inner_data

        if self.attr_specs.is_list:
            pass


        for key, value in widget.get_data().items():
            data[key] = value

        if self.attr_specs.is_list:
            inner_data = {}
            for key, value in widget.get_data().items():
                inner_data[key] = value
            data_list.append(inner_data)
        return

    def _process_input_panel_wrapper(self, widget, data_list, data):
        if self.attr_specs.is_list:
            data_list.append(widget.get_input_value())
        inner_data = widget.get_input_value()

        if widget.get_short_attr_name() in inner_data:
            for item_key, item_value in inner_data.items():
                data[item_key] = item_value
            return

        data[widget.get_short_attr_name()] = inner_data

    def get_data(self):
        """Group and gather all the data from the input widget."""

        data = {}
        data_list = []

        for i in range(self.layout().count()):
            item = self.layout().itemAt(i)
            if item:
                widget = item.widget()

                if widget:
                    if isinstance(widget, QLineEdit):
                        data[self.attr_name] = widget.text()

                        if self.attr_specs.is_list:
                            data_list.append(widget.text())
                        continue

                    if isinstance(widget, BlockTypeWrapper):


                        self._process_block_type_input(widget, data_list, data)


                    if isinstance(widget, BlockItemWrapper):


                        self._process_block_item_type_input(widget, data_list, data)


                    if widget.__class__.__name__ == "InputPanelWrapper":
                        self._process_input_panel_wrapper(widget, data_list, data)

        if self.attr_specs.is_list:
            return data_list


        first_key = next(iter(data))

        list_of_dicts = data[first_key]
        if not isinstance(list_of_dicts, list):
            return data
        if not self.attr_specs.has_nested_attributes:
            return data

        for attr_name, attr_spec in self.attr_specs.nested_attributes.items():
            if attr_spec.is_list:
                continue
            single_dict = {k: v for d in list_of_dicts for k, v in d.items()}
            data[first_key] = single_dict

        return data



import re


def process_key(s):
    result = s[s.rfind("."):] if s.__contains__(".") else s
    return result

def convert_to_int(s):
    # Extract the first sequence of digits from the string
    match = re.search(r'\d+', s)
    if match:
        return int(match.group())  # Convert the matched digits to an integer
    else:
        raise ValueError("No valid integer found in the string")

def extract_id_from_contract_identifier(contract_identifier: str) -> int:
    index = contract_identifier.find('#')
    after_find = contract_identifier[index + 1:].strip() if index != -1 else None

    return convert_to_int(after_find)


def extract_product_name_from_contract_identifier(contract_identifier: str):
    parts = contract_identifier.split('(', 1)  # Split at first '('
    if len(parts) > 1:
        return parts[1].split(')', 1)[0]  # Extract the part before the first ')'
    return ""  # Return empty string if no parentheses are found


def get_last_substring(s):
    """
    Returns the substring after the last '.' in the input string.
    If no '.' is present, returns the original string.
    """
    return s.rsplit('.', 1)[-1]



import logging

import ujson as json
import os
from pathlib import Path

from PySide6.QtCore import QObject

from famegui_runtime_helper.dict_hash_helper import hash_dict
from famegui_runtime_helper.fame_session_cache import FameSessionCache
from famegui.config.config import BASE_DIR
from famegui.path_config import CACHE_TRACKING_DB_PATH, CACHE_TRACKING_DB_DIR
from famegui.undo_utils.contract_utils import ORIGINAL_CONTRACT_KEY, MANIPULATED_CONTRACT_KEY
from famegui.undo_utils.undo_constants import UndoActions


def bulk_save_old_agent_coords(agent_ctrl_dict: dict,
                               agent_ctrl_dict_post_transformational_coords: dict):
    with open(CACHE_TRACKING_DB_PATH, "r", encoding="utf-8") as file:
        json_db_data = json.loads(file.read())

    data_list = json_db_data["data_list"]

    if hash_dict(agent_ctrl_dict_post_transformational_coords) == hash_dict(agent_ctrl_dict):
        raise ValueError("Pre and Post-Transformational Coords cannot be the same !")

    if len(data_list) >= 2:
        first_item = data_list[0]
        last_item = data_list[len(data_list) - 1]
        item_idx = first_item["index"]
        item_idx_last = last_item["index"]

        if item_idx > item_idx_last:
            last_item_in_list_by_idx = first_item
        else:
            last_item_in_list_by_idx = last_item
        last_item_idx = last_item_in_list_by_idx["index"]

        new_idx = last_item_idx + 1

    else:
        new_idx = len(data_list) + 1

    undo_step = {
        "type": "agent_re_arrange",
        "data": {
            "agent_coords": agent_ctrl_dict,
            "agent_ctrl_dict_post_transformational_coords": agent_ctrl_dict_post_transformational_coords
        },
        "index": new_idx
    }

    session_db = FameSessionCache()

    session_db.add_item(undo_step)

    data_list.append(undo_step)

    with open(CACHE_TRACKING_DB_PATH, "w", encoding="utf-8") as file:
        file.write(json.dumps(json_db_data))


class UndoManager(QObject):

    def __init__(self, controller):
        super().__init__()
        self.check_or_create_cache_db()

        self.session_db = FameSessionCache()

        self.controller = controller

    @staticmethod
    def check_or_create_cache_db():
        json_db_data = {"data_list": [], "current_index": 0}

        os.makedirs(CACHE_TRACKING_DB_DIR, exist_ok=True)

        file_path = Path(CACHE_TRACKING_DB_PATH)

        if not file_path.is_file():
            with open(CACHE_TRACKING_DB_PATH, "w+", encoding="utf-8") as file:
                file.write(json.dumps(json_db_data))

    def get_current_cache_index(self):
        cache_db = os.path.join(BASE_DIR, "cache", "db", "db.json")

        data_dict = {}

        with open(cache_db, "r", encoding="utf-8") as db_file:
            data_dict = json.loads(db_file.read())

        return data_dict["current_index"]

    def add_new_undo_step(self, data_dict):
        cache_db = os.path.join(BASE_DIR, "cache", "db", "db.json")

        current_db = {}

        with open(cache_db, "r", encoding="utf-8") as read_db:
            current_db = json.loads(read_db.read())

        data_list = current_db["data_list"]
        data_list.append(data_dict)

        self.session_db.add_item(data_dict)

        with open(cache_db, "w+", encoding="utf-8") as db_file:
            db_file.write(json.dumps(current_db))

    def save_project_properties_manipulation(self, original_props: dict, updated_props: dict):

        data = {
            "type": UndoActions.UNDO_SCENARIO_PROPERTIES.value,
            "data": {
                "original_props": original_props,  # use for undo
                "updated_props": updated_props,  # use for redo
            },
            "index": self.get_current_cache_index() + 1
        }

        self.add_new_undo_step(data)

    def save_moved_agent(self, agent_x, agent_y, agent_id, old_x, old_y):
        data = {
            "type": "agent_moved",
            "data": {
                "agent": {
                    "agent_x": agent_x,
                    "agent_y": agent_y,
                    "agent_id": agent_id
                },
                "original_agent": {
                    "agent_x": old_x,
                    "agent_y": old_y,
                    "agent_id": agent_id

                }
            },
            "index": self.get_current_cache_index() + 1
        }

        self.add_new_undo_step(data)

    def peek_last_cache_file(self):
        # List all files in the directory
        result = None
        cache_db = os.path.join(BASE_DIR, "cache", "db", "db.json")

        with open(cache_db, "r", encoding="utf-8") as f:
            data = json.load(f)
            current_index = int(data["current_index"])
            data_list = data["data_list"]
            for item in data_list:
                if item["index"] == len(data_list) - current_index:
                    result = item
                    break

        return result["file_name"]

    def perform_last_queued_redo(self):
        """Performing redo by inverting the undo actions the """
        result = self.session_db.redo_last_item()
        logging.info("Redo inside ....")
        logging.info(result)

        if result is None:
            return

        if result["type"] == UndoActions.UNDO_CONTRACT_FIELDS_CHANGED.value:
            contract_data = result["data"]

            original_contract = contract_data[MANIPULATED_CONTRACT_KEY]
            current_contract = contract_data[ORIGINAL_CONTRACT_KEY]

            self.controller.perform_contract_fields_undo(current_contract, original_contract)

        if result["type"] == UndoActions.UNDO_AGENT_ATTRIBUTES_CHANGED.value:
            self.controller.perform_agent_attributes_undo(result["data"])

        if result["type"] == "agent_re_arrange":
            agent_coords_data = result["data"]["agent_ctrl_dict_post_transformational_coords"]

            self.controller.re_render_scene(agent_coords_data)

        if result["type"] == UndoActions.UNDO_SCENARIO_PROPERTIES.value:
            self.controller.apply_properties(result["data"]["updated_props"])

        if result["type"] == "agent_moved":
            agent_data = result["data"]["original_agent"]

            self.controller.set_agent_display_xy(agent_data["agent_id"], agent_data["agent_x"],
                                                 agent_data["agent_y"])

    def perform_last_queued_undo(self):
        result = self.session_db.get_last_item()

        if result is None:
            return

        if result["type"] == UndoActions.UNDO_CONTRACT_FIELDS_CHANGED.value:
            contract_data = result["data"]

            original_contract = contract_data[ORIGINAL_CONTRACT_KEY]
            current_contract = contract_data[MANIPULATED_CONTRACT_KEY]

            self.controller.perform_contract_fields_undo(current_contract, original_contract)

        if result["type"] == UndoActions.UNDO_AGENT_ATTRIBUTES_CHANGED.value:
            self.controller.perform_agent_attributes_undo(result["data"])

        if result["type"] == UndoActions.UNDO_SCENARIO_PROPERTIES.value:
            self.controller.apply_properties(result["data"]["original_props"])

        if result["type"] == "agent_re_arrange":
            agent_coords_data = result["data"]["agent_coords"]

            self.controller.re_render_scene(agent_coords_data)

        if result["type"] == "agent_moved":
            agent_data = result["data"]["agent"]

            self.controller.set_agent_display_xy(agent_data["agent_id"], agent_data["agent_x"],
                                                 agent_data["agent_y"])

    def perform_undo(self):

        self.perform_last_queued_undo()

    def perform_redo(self):

        self.perform_last_queued_redo()



from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtWidgets import QDialog, QDialogButtonBox, QHBoxLayout

from famegui import models
from famegui.agent_controller import AgentController
from famegui.config.config import CUSTOM_DELIVERY_STEPS_TIME_OPTION
from famegui.config.contract_consts import PRESET_DURATION_IN_SECONDS
from famegui.config.static_ui_components import build_new_contract_dlg
from famegui.config.style_config import FAME_LINE_EDIT_STYLE_ERR_STATE, FAME_LINE_EDIT_STYLE_DEFAULT, \
    ERROR_TEXT_STYLE_LABEL
from famegui.database.prebuild_queries.interval_presets import get_delivery_interval_presets
from famegui.generated.ui_dialog_newcontract import Ui_DialogNewContract
from famegui.time_utils import (
    convert_fame_time_to_gui_datetime,
    convert_gui_datetime_to_fame_time,
)
from famegui.ui.helper import gen_preset_interval_chooser


class DialogNewContract(QDialog):
    """Dialog for creating a single contract"""

    fame_to_datetime_conversion_triggered: bool = False
    WIDGET_DATE_FORMAT = "dd/MM/yyyy hh:mm:ss"

    def _configure_line_edit_for_unsigned_int(self, line_edit: QtWidgets.QLineEdit):
        line_edit.setText("0")
        regex_uint64 = QtCore.QRegularExpression("\\d{1,20}")
        line_edit.setValidator(QtGui.QRegularExpressionValidator(regex_uint64))
        line_edit.textChanged.connect(self._update_ok_button_status)

    def on_preset_changed(self, index, selected_option: str):
        """
        Handles the change event when a new preset is selected
        """
        # Do something with the selected preset
        if selected_option == "Custom":
            return

        value = self.interval_presets[index][PRESET_DURATION_IN_SECONDS]

        self._ui.lineDeliveryInterval.setText(value)


    def _configure_line_edit_for_signed_int(self, line_edit: QtWidgets.QLineEdit):
        line_edit.setText("0")
        regex_uint64 = QtCore.QRegularExpression("-?\\d{1,20}")
        line_edit.setValidator(QtGui.QRegularExpressionValidator(regex_uint64))
        line_edit.setStyleSheet("background-color: #EBEBEB;")
        line_edit.textChanged.connect(self._update_ok_button_status)

    def gen_preset_interval_chooser(self):
        """
        Generates a simple ComboBox/Preset chooser Widget to choose from presets
        """

        choices = [item["preset_label"] for item in self.interval_presets]
        choices_bundles = [item for item in self.interval_presets]

        interval_preset_chooser = gen_preset_interval_chooser(choices, self.on_preset_changed,
                                                              self._ui.lineDeliveryInterval,
                                                              choices_bundles,
                                                              CUSTOM_DELIVERY_STEPS_TIME_OPTION)

        return interval_preset_chooser

    def _setup_delivery_interval_chooser_input(self):
        """
        Insert a simple ComboBox/Preset chooser Widget
        into the delivery interval chooser box to choose from Interval presets
        """

        chooser = self.gen_preset_interval_chooser()

        chooser_holder_layout = QHBoxLayout()

        chooser_holder_layout.addWidget(chooser)
        chooser_holder_layout.setContentsMargins(0, 0, 0, 0)

        self._ui.delivery_interval_chooser.setLayout(chooser_holder_layout)

    def _configure_line_edit_for_optional_signed_int(self, line_edit: QtWidgets.QLineEdit):
        line_edit.setText("")
        line_edit.setPlaceholderText(self.tr("optional"))
        regex_uint64 = QtCore.QRegularExpression("-?\\d{0,20}")
        line_edit.setValidator(QtGui.QRegularExpressionValidator(regex_uint64))
        line_edit.textChanged.connect(self._update_ok_button_status)

    def __init__(
            self,
            sender: AgentController,
            receiver: AgentController,
            schema: models.Schema,
            parent=None,
    ):
        QDialog.__init__(self, parent)
        self._ui = Ui_DialogNewContract()
        self._ui.setupUi(self)
        self._sender = sender
        self._receiver = receiver

        self.interval_presets = get_delivery_interval_presets()

        self.setWindowTitle(self.tr("New contract"))
        self._ui.labelDescr.setText(
            self.tr(self.tr(self.tr(build_new_contract_dlg(self._sender, self._receiver)))))

        # force the user to select a product except if only one is available
        if self._ui.comboBoxProduct.count() != 1:
            self._ui.comboBoxProduct.setCurrentIndex(-1)

        self._ui.comboBoxProduct.currentIndexChanged.connect(self._update_ok_button_status)
        self._update_ok_button_status()

        self._prep_input_line_edits()

        # fill possible products to select based on the sender schema agent type
        sender_type = schema.agent_type_from_name(sender.type_name)
        assert sender_type is not None
        self._ui.comboBoxProduct.addItems(list(set(sender_type.products)))

        # force the user to select a product except if only one is available
        if self._ui.comboBoxProduct.count() != 1:
            self._ui.comboBoxProduct.setCurrentIndex(-1)

        # connect
        self._ui.comboBoxProduct.currentIndexChanged.connect(
            self._update_ok_button_status
        )

        self._update_ok_button_status()

        self._ui.lineFirstDeliveryTime.textChanged.connect(
            self._update_fame_time_text_areas
        )

        self._ui.lineExpirationTime.textChanged.connect(
            self._update_fame_time_text_areas
        )

    def accept(self):
        delivery_fame_time = int(self._ui.lineDeliveryInterval.text())

        if delivery_fame_time == 0:
            self._ui.lineDeliveryInterval.setStyleSheet(FAME_LINE_EDIT_STYLE_ERR_STATE)
            self._ui.labelErrorDeliveryInterval.setText("Delivery Interval must be greater then 0!")

            return

        self._ui.lineDeliveryInterval.setStyleSheet(FAME_LINE_EDIT_STYLE_DEFAULT)
        self._ui.labelErrorDeliveryInterval.setText("")

        super().accept()

    def _prep_date_time_fields(self):
        """Configure date time fields to use the correct format and to trigger the update of the fame time fields"""
        self._ui.lineFirstDeliveryNonFameTime.setCalendarPopup(True)
        self._ui.lineFirstDeliveryNonFameTime.setDisplayFormat(self.WIDGET_DATE_FORMAT)
        self._ui.lineFirstDeliveryNonFameTime.dateTimeChanged.connect(
            self._update_fame_times
        )

        self._ui.lineExpirationTimeNonFameTime.setCalendarPopup(True)
        self._ui.lineExpirationTimeNonFameTime.setDisplayFormat(self.WIDGET_DATE_FORMAT)
        self._ui.lineExpirationTimeNonFameTime.dateTimeChanged.connect(
            self._update_fame_times
        )

    def _prep_input_line_edits(self):
        """Accept uint64 numbers as specified in protobuf schema"""
        self._configure_line_edit_for_unsigned_int(self._ui.lineDeliveryInterval)
        self._configure_line_edit_for_signed_int(self._ui.lineFirstDeliveryTime)
        self._configure_line_edit_for_optional_signed_int(self._ui.lineExpirationTime)

        self._setup_delivery_interval_chooser_input()

    def make_new_contract(self) -> models.Contract:
        expiration_time_str = self._ui.lineExpirationTime.text()
        expiration_time = int(expiration_time_str) if expiration_time_str else None

        return models.Contract(
            self._sender.id,
            self._receiver.id,
            self._ui.comboBoxProduct.currentText(),
            int(self._ui.lineDeliveryInterval.text()),
            int(self._ui.lineFirstDeliveryTime.text()),
            expiration_time,
        )

    def _update_fame_time_text_areas(self):
        self.fame_to_datetime_conversion_triggered = True

        fame_first_delivery_time: str = self._ui.lineFirstDeliveryTime.text()
        if fame_first_delivery_time == "":
            return
        gui_start_time = convert_fame_time_to_gui_datetime(
            int(fame_first_delivery_time))
        self._ui.lineFirstDeliveryNonFameTime.setDateTime(gui_start_time)

        fame_expiration_time: str = self._ui.lineExpirationTime.text()
        if fame_expiration_time == "":
            return

        gui_expiration_time_time = convert_fame_time_to_gui_datetime(int(fame_expiration_time))

        self._ui.lineExpirationTimeNonFameTime.setDateTime(gui_expiration_time_time)

    def _update_fame_times(self):
        if self.fame_to_datetime_conversion_triggered:
            self.fame_to_datetime_conversion_triggered = False
            return
        self._ui.lineExpirationTime.setText(
            str(convert_gui_datetime_to_fame_time(
                self._ui.lineExpirationTimeNonFameTime.text()
            )))

        self._ui.lineFirstDeliveryTime.setText(
            str(
                convert_gui_datetime_to_fame_time(
                    self._ui.lineFirstDeliveryNonFameTime.text()
                )))

    def _update_ok_button_status(self):
        all_fields_ok = (
                self._ui.comboBoxProduct.currentText() != ""
                and self._ui.lineDeliveryInterval.text() != ""
                and self._ui.lineFirstDeliveryTime.text() != ""
        )
        self._ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(all_fields_ok)



import os
from copy import deepcopy

from fameio.input.resolver import PathResolver

from famegui_runtime_helper.attribute_dict_formatting_helper import build_simple_attr_dict
from famegui_runtime_helper.dict_hash_helper import hash_dict
from famegui import models
from famegui.config.config import DEFAULT_FAME_WORK_DIR
from famegui.models.tests.test_utils.test_utils import replace_first_number



def test_all_available_agents_for_update():
    assert os.path.exists(DEFAULT_FAME_WORK_DIR), f"Path does not exist: {DEFAULT_FAME_WORK_DIR}"

    file_path = os.path.join(DEFAULT_FAME_WORK_DIR, "scenario.yaml")

    scenario_model = models.ScenarioLoader.load_yaml_file(
        file_path, PathResolver(), DEFAULT_FAME_WORK_DIR)

    assert scenario_model is not None

    already_fetched_agent_types = set()

    first_agent_type_matches = []

    for agent in scenario_model.agents:

        if already_fetched_agent_types.__contains__(agent.type_name):
            continue

        first_agent_type_matches.append(agent)

        already_fetched_agent_types.add(agent.type_name)

    first_agent_type_matches = [agent for agent in first_agent_type_matches if
                                agent.attributes and scenario_model.schema.agent_types[
                                    agent.type_name
                                ].attributes]

    for item in first_agent_type_matches:
        flat_data = item.get_flat_agent_attributes_dict()

        original_data = deepcopy(flat_data)

        updated_flat_data = deepcopy(flat_data)

        agent_type_name = item.type_name

        data_out = {}

        updated_flat_data, _ = replace_first_number(updated_flat_data)

        assert hash_dict(updated_flat_data) != hash_dict(
            original_data)  # check if test is able to replace primitive value

        for attr_name, attr in scenario_model.schema.agent_types[agent_type_name].attributes.items():
            iter_tmp_dict = build_simple_attr_dict(
                attr, updated_flat_data, agent_type_name)
            data_out[attr_name] = iter_tmp_dict[attr_name] if attr_name in iter_tmp_dict else iter_tmp_dict

        # test undo

        for attr_name, attr in scenario_model.schema.agent_types[agent_type_name].attributes.items():
            iter_tmp_dict = build_simple_attr_dict(
                attr, original_data, agent_type_name)
            data_out[attr_name] = iter_tmp_dict[attr_name] if attr_name in iter_tmp_dict else iter_tmp_dict

        newest_attr_data = item.get_flat_agent_attributes_dict()

        assert hash_dict(newest_attr_data) == hash_dict(
            original_data)



import os
from copy import deepcopy

from fameio.input.resolver import PathResolver

from famegui_runtime_helper.attribute_dict_formatting_helper import build_simple_attr_dict
from famegui_runtime_helper.dict_hash_helper import hash_dict
from famegui import models
from famegui.config.config import DEFAULT_FAME_WORK_DIR
from famegui.fame_tools.dict_tools import remove_none_values_and_empty_elements


def test_simple_attribute_update():
    assert os.path.exists(DEFAULT_FAME_WORK_DIR), f"Path does not exist: {DEFAULT_FAME_WORK_DIR}"

    file_path = os.path.join(DEFAULT_FAME_WORK_DIR, "scenario.yaml")

    scenario_model = models.ScenarioLoader.load_yaml_file(
        file_path, PathResolver(), DEFAULT_FAME_WORK_DIR)

    assert scenario_model is not None

    first_agent = deepcopy(scenario_model.agents[0])

    data_out = {}

    flat_full_name_labeled_values = first_agent.get_flat_agent_attributes_dict()

    flat_full_name_labeled_values_updated = deepcopy(flat_full_name_labeled_values)

    flat_full_name_labeled_values_updated["GateClosureInfoOffsetInSeconds"] = 29

    agent_type_name = first_agent.type_name

    for attr_name, attr in scenario_model.schema.agent_types[agent_type_name].attributes.items():
        iter_tmp_dict = build_simple_attr_dict(
            attr, flat_full_name_labeled_values_updated, agent_type_name)
        data_out[attr_name] = iter_tmp_dict[attr_name] if attr_name in iter_tmp_dict else iter_tmp_dict

    data_out = remove_none_values_and_empty_elements(data_out)

    for attr_name_sub, attr_sub in data_out.items():
        value = data_out[attr_name_sub]
        first_agent.attributes[attr_name_sub] = value

    scenario_model.update_agent(first_agent)

    new_flat_full_name_labeled_values = first_agent.get_flat_agent_attributes_dict()

    # agent attributes must be changed

    assert hash_dict(flat_full_name_labeled_values) != hash_dict(new_flat_full_name_labeled_values)



import typing

from famegui.database.db_loader.db_manager import get_rows
from famegui.database.settings_db.init_db import init_db_session


def get_delivery_interval_presets() -> typing.List[typing.Dict]:
    """
        Retrieve all delivery interval presets from the 'delivery_interval_presets' table.

        This function uses an active database session to query the 'delivery_interval_presets'
        table and returns a list of dictionaries containing the 'preset_duration_in_seconds'
        and 'preset_label' fields for each row.

        Returns:
            List[Dict[str, Any]]: A list of dictionaries with 'preset_duration_in_seconds'
                                  and 'preset_label' for each delivery interval preset.
        """
    with init_db_session() as session:
        res_list = get_rows(session, "delivery_interval_presets", ["preset_duration_in_seconds", "preset_label"])



    return res_list



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'delete_agent_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *


class Ui_DeleteAgentDialog(object):
    def setupUi(self, DeleteAgentDialog):
        if not DeleteAgentDialog.objectName():
            DeleteAgentDialog.setObjectName("DeleteAgentDialog")
        DeleteAgentDialog.resize(400, 300)
        self.verticalLayout = QVBoxLayout(DeleteAgentDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.treeWidget = QTreeWidget(DeleteAgentDialog)
        __qtreewidgetitem = QTreeWidgetItem()
        __qtreewidgetitem.setText(0, "1")
        self.treeWidget.setHeaderItem(__qtreewidgetitem)
        self.treeWidget.setObjectName("treeWidget")

        self.verticalLayout.addWidget(self.treeWidget)

        self.checkBoxDeleteAgent = QCheckBox(DeleteAgentDialog)
        self.checkBoxDeleteAgent.setObjectName("checkBoxDeleteAgent")
        self.checkBoxDeleteAgent.setText("Delete all related Contracts")

        self.verticalLayout.addWidget(self.checkBoxDeleteAgent)

        self.checkBoxDeleteAllContracts = QCheckBox(DeleteAgentDialog)
        self.checkBoxDeleteAllContracts.setObjectName("checkBoxDeleteAllContracts")
        self.checkBoxDeleteAllContracts.setText("Delete all related Contracts")

        self.verticalLayout.addWidget(self.checkBoxDeleteAllContracts)

        self.checkBoxDeleteAllConnectedAgents = QCheckBox(DeleteAgentDialog)
        self.checkBoxDeleteAllConnectedAgents.setObjectName("checkBoxDeleteAllConnectedAgents")
        self.checkBoxDeleteAllConnectedAgents.setText("Delete all related Agents")

        self.verticalLayout.addWidget(self.checkBoxDeleteAllConnectedAgents)

        self.retranslateUi(DeleteAgentDialog)

        QMetaObject.connectSlotsByName(DeleteAgentDialog)

    # setupUi

    def retranslateUi(self, DeleteAgentDialog):
        DeleteAgentDialog.setWindowTitle(QCoreApplication.translate("DeleteAgentDialog", "Delete Agent Dialog", None))

    # retranslateUi



import fameio.input.schema as schema

INPUT_ATTR_TYPE_COLORS = {
    schema.AttributeType.BLOCK: (255, 221, 148, 0.7),
    schema.AttributeType.LONG: (0, 188, 180, 200),
    schema.AttributeType.INTEGER: (255, 204, 75, 200),
    schema.AttributeType.ENUM: (255, 231, 32, 200),
    schema.AttributeType.DOUBLE: (39, 78, 221, 200),
    schema.AttributeType.TIME_SERIES: (255, 169, 33, 200),
}

DEFAULT_COLOR = (39, 78, 221, 200)
BORDER = "border: 1.5px {};"
BORDER_MANDATORY = "solid black"
BORDER_STANDARD = "dashed #141414"

ERROR_TEXT_STYLE_LABEL = "QLabel { background-color : #F66257; }"

ERROR_TEXT_STYLE_LINE_EDIT = "QLineEdit { background-color : #F66257; }"

FAME_CALENDAR_INPUT_STYLE = "QLineEdit { background: '#7BDFF2'; }"

FAME_LINE_EDIT_STYLE_DEFAULT = "background-color: #f5f5f5;"

FAME_LINE_EDIT_STYLE_ERR_STATE = ("QWidget{ background-color: rgba(255, 102, 102, 150);"
                                  " border: 2px dashed red; border-radius: 5px; }")

NON_SUPPORTED_LABEL_STYLE = "QLabel { background-color : #F66257; };"

FAME_CONSOLE_STYLE = "color: #F5F5F5;"

FAME_CONSOLE_TEXT_STYLE = """
            QTextEdit {
                background-color: #141414;  /* Light gray background */
                color: #F0F0F0;  /* Black text */
            }
        """

LIST_ITEM_STYLE = "background-color: rgba(255, 221, 148, 0.4); border-radius: 5px;"

DEFAULT_INPUT_STYLE = ("QTextEdit { padding-left: 20px; padding-right: 20px; background-color: #f5f5f5; "
                       "border: 1px solid #d0d0d0; border-radius: 5px; padding: 10px; color: #333; "
                       "font-size: 12px; font-family: Arial; } QTextEdit:focus { border: 1px solid #0078d4; }"
                       " QTextEdit QScrollBar:vertical { border: none; background: #f5f5f5; width: 10px; margin: 0; }"
                       " QTextEdit QScrollBar::handle:vertical { background: #d0d0d0; min-height: 20px; }"
                       " QTextEdit QScrollBar::add-line:vertical,"
                       " QTextEdit QScrollBar::sub-line:vertical { border: none; background: none; }"
                       " QTextEdit QScrollBar:horizontal { border: none; background: #f5f5f5; height: 10px; margin: 0; }"
                       " QTextEdit QScrollBar::handle:horizontal { background: #d0d0d0; min-width: 20px; }"
                       " QTextEdit QScrollBar::add-line:horizontal,"
                       " QTextEdit QScrollBar::sub-line:horizontal { border: none; background: none; }")

REMOVE_BTN_STYLE = ("QPushButton { background-color: #D32F2F; color: white; "
                    "border: none; padding: 10px; border-radius: 5px; } "
                    "QPushButton:hover { background-color: #F44336; }"
                    " QPushButton:pressed { background-color: #B71C1C; }")

COLLAPSE_BTN_STYLE = (
    "QPushButton { background-color: #607D8B; color: white; border: none;"
    " padding: 10px; border-radius: 5px; } "
    "QPushButton:hover { background-color: #78909C; }"
    " QPushButton:pressed { background-color: #455A64; }")


def get_color_for_type(attr_spec) -> str:
    """Returns the background color for the attribute type"""
    color = INPUT_ATTR_TYPE_COLORS.get(attr_spec.attr_type, DEFAULT_COLOR)
    return f"rgba{color}"


def get_default_input_style(uniq_widget_object_name: str) -> str:
    return \
        f"""#{uniq_widget_object_name} 
        {{ background-color: #FAFAFA;
         border: 1px solid grey; border-radius: 5px; }}"""


def get_default_err_input_style(uniq_widget_object_name: str) -> str:
    return \
        (f"#{uniq_widget_object_name} "
         f"{{ background-color: rgba(255, 102, 102, 150);"
         f" border: 1px solid red; border-radius: 5px; }}")


def get_border_for_type(attr_spec: schema.AttributeSpecs) -> str:
    """Returns border style depending on type"""
    border_style = BORDER_MANDATORY if attr_spec.is_mandatory else BORDER_STANDARD
    return BORDER.format(border_style)



from typing import Union

from famegui_runtime_helper.dict_hash_helper import hash_dict
from famegui.models import Scenario, Contract

MANIPULATED_CONTRACT_KEY = "current_contract"
ORIGINAL_CONTRACT_KEY = "original_contract"


# noinspection PyTypeChecker
def swap_or_delete_contract(scenario: Scenario, target_ref_contract: Union[dict, None],
                            contract_to_be_replaced: Union[dict, None]):
    if target_ref_contract is None:
        scenario.contracts.append(Contract.from_dict(contract_to_be_replaced))
        return

    item_idx_to_manipulate = None

    for idx, contract_item in enumerate(scenario.contracts):

        if hash_dict(contract_item.to_dict()) == hash_dict(target_ref_contract):
            item_idx_to_manipulate = idx
            break


    if contract_to_be_replaced is None:
        scenario.contracts.remove(scenario.contracts[item_idx_to_manipulate])
        return

    assert item_idx_to_manipulate is not None

    scenario.contracts[item_idx_to_manipulate] = Contract.from_dict(contract_to_be_replaced)



import logging
import os.path
from contextlib import contextmanager

from sqlalchemy import create_engine, Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from famegui.config.config import BASE_DIR
from typing import Union, List, overload

# Define the database URI
DATABASE_URI = f'sqlite:///{os.path.join(BASE_DIR, "database.sqlite3")}'

# Create the engine
engine = create_engine(DATABASE_URI)

# Define the base class
Base = declarative_base()

# Create a session factory
Session = sessionmaker(bind=engine)


class Setting(Base):
    __tablename__ = 'settings'

    id = Column(Integer, primary_key=True, autoincrement=True, unique=True)
    value = Column(String, nullable=False)
    setting_name = Column(String, nullable=False)
    data_type = Column(String, nullable=False)  # Adding the data_type column
    setting_label = Column(String, nullable=False)
    help_text = Column(String, nullable=True)

    def __repr__(self):
        return f"<Setting(id={self.id}, value='{self.setting_name}')>"


def get_db_engine():
    engine = create_engine(DATABASE_URI)
    return engine





@contextmanager
def init_db_session():
    # Create the table if it does not exist
    Base.metadata.create_all(engine)

    # Create a session
    session = Session()

    try:
        yield session
    finally:
        # Close the session after use
        session.close()







if __name__ == '__main__':
    logging.info("Running init_db.py")




import os.path
from copy import deepcopy

from fameio.input.resolver import PathResolver
from fameio.input.scenario import StringSet

from famegui.config.config import DEFAULT_FAME_WORK_DIR
from famegui.models import ScenarioLoader, Scenario


def get_default_scenario_model():
    """ Helper function to retrieve base scenario
    model obj for prototyping and testing"""

    scenario_model = ScenarioLoader.load_yaml_file(
        os.path.join(DEFAULT_FAME_WORK_DIR, "scenario.yaml"),
        PathResolver(),
        DEFAULT_FAME_WORK_DIR
    )

    return scenario_model


def add_string_set_value(scenario: Scenario, string_set_group_name: str, string_set_value: str):
    item = scenario.string_sets[string_set_group_name]

    value_dict: dict = item.to_dict()

    value_dict_copy = deepcopy(value_dict["values"])
    value_dict_copy[string_set_value] = {}

    value_dict["values"].update(value_dict_copy)
    data = StringSet.from_dict(value_dict)

    scenario.string_sets[string_set_group_name] = data



if __name__ == "__main__":
    model = get_default_scenario_model()

    ScenarioLoader.save_to_yaml_file(model, os.path.join(DEFAULT_FAME_WORK_DIR, "scenario.yaml"))



from PySide6.QtWidgets import QWidget, QVBoxLayout, QPushButton, QLineEdit, QListWidget


def get_list_widget_holder():
    """
    Create and return a QWidget containing a QListWidget, an input field, and buttons to add/remove items.
    """

    holder = QWidget()
    layout = QVBoxLayout()
    lineEdit = QLineEdit(holder)

    layout.addWidget(lineEdit)

    addButton = QPushButton("Add", holder)
    layout.addWidget(addButton)

    listWidget = QListWidget(holder)
    layout.addWidget(listWidget)

    removeButton = QPushButton("Remove Selected", holder)
    layout.addWidget(removeButton)

    holder.setLayout(layout)

    return holder



from setuptools import setup, find_packages
import os


def _load_famegui_version():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    init_filepath = os.path.join(this_dir, "famegui", "__init__.py")

    with open(init_filepath) as f:
        for line in f.readlines():
            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise RuntimeError("Unable to find version string in {}".format(init_filepath))


def _readme():
    with open("README.md") as f:
        return f.read()


VERSION = _load_famegui_version()

setup(
    name="famegui",
    packages=find_packages(),
    version=VERSION,
    keywords=["FAME", "agent-based modelling"],
    license="Apache License 2.0",
    description="Graphical user interface to the FAME modelling framework",
    long_description=_readme(),
    long_description_content_type="text/markdown",
    author="Aurélien Regat-Barrel, Simon Wischnevetski",
    author_email="fame@dlr.de",

    url="https://gitlab.com/fame-framework/FAME-Gui",
    download_url="https://gitlab.com/fame-framework/FAME-Gui/-/archive/v{}/FAME-Gui-v{}.tar.gz".format(
        VERSION, VERSION
    ),
    install_requires=[
        "coloredlogs",
        "fameio>=3.0",
        "python-igraph",
        "pyyaml",
        "ipython",
        "PySide6",
        "SQLAlchemy>=1.4.44",
        "pytest==8.3.4",
        "matplotlib",
        "ujson",
        "icecream",
    ],
    python_requires=">=3.9",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Visualization",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    entry_points={
        "console_scripts": [
            "famegui=famegui.app:main",
        ],
    },
    package_data={
        "famegui": ["data/*", "database/database.sqlite3", "*sqlite3"],
    },
    include_package_data=True,
)



import os

import fameio.input.scenario as fameio
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import QSize, Qt, QEvent, QObject
from PySide6.QtUiTools import QUiLoader
from PySide6.QtWidgets import QPushButton

from famegui import models
from famegui.appworkingdir import AppWorkingDir
from famegui.ui.attribute_row_item import AttributeTreeItem

import fameio.input.schema as schema

from famegui.ui.quick_modals import gen_quick_warning_modal
from famegui_runtime_helper.attribute_dict_formatting_helper import get_sub_attr


class DialogNewAgent(QtWidgets.QDialog):
    accepted = QtCore.Signal()
    trigger_input_validation = QtCore.Signal()
    validation_error_signal = QtCore.Signal(bool, str)

    def __init__(
            self,
            schema: models.Schema,
            working_dir: AppWorkingDir,
            scenario: models.Scenario,
            parent=None,
    ):
        super().__init__()

        loader = QUiLoader()
        file = QtCore.QFile(
            os.path.join(os.path.dirname(__file__), "dialog_newagent.ui")
        )

        file.open(QtCore.QFile.OpenModeFlag.ReadOnly)

        self._ui = loader.load(file, self)
        self._ui.setParent(None)
        self.scenario: models.Scenario = scenario

        file.close()

        self._schema = schema
        self._working_dir = working_dir
        self._tree_items = []

        self.validation_data = []

        self.setup_ui()
        self.connect_slots()

        # connect slots

    def setup_ui(self):
        self.fired_once = False
        self.saved_data = None

        # init
        self.setWindowTitle(self.tr("New agent"))
        self._ui.comboBoxType.addItems(self._schema.agent_types.keys())

        self._ui.setWindowFlags(
            Qt.Window | Qt.WindowCloseButtonHint | Qt.WindowMaximizeButtonHint
        )

        # tree view
        self._ui.treeWidget.setSelectionMode(
            QtWidgets.QAbstractItemView.SingleSelection
        )
        self._ui.treeWidget.setRootIsDecorated(False)
        self._ui.treeWidget.setColumnCount(2)
        self._ui.treeWidget.setHeaderLabels([self.tr("Attribute"), self.tr("Value")])
        self._ui.treeWidget.setColumnWidth(0, 200)
        self._ui.treeWidget.setAlternatingRowColors(True)
        self._reset_attributes()

        self.setWindowFlag(Qt.WindowMinimizeButtonHint, True)
        self.setWindowFlag(Qt.WindowMaximizeButtonHint, True)
        flags = self.windowFlags()
        flags |= Qt.CustomizeWindowHint
        flags |= Qt.WindowTitleHint
        flags |= Qt.WindowSystemMenuHint
        flags |= Qt.WindowCloseButtonHint
        flags |= Qt.WindowStaysOnTopHint
        flags |= Qt.WindowMaximizeButtonHint
        self.setWindowFlags(flags)

    def connect_slots(self):
        """Used to handle user input and to prevent default behavior of the dialog"""
        self._ui.buttonBox.accepted.disconnect()

        self._ui.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).clicked.connect(self.on_ok_clicked)

        self._ui.comboBoxType.currentTextChanged.connect(self._reset_attributes)
        self.validation_error_signal.connect(self.on_err_validation)

    def on_err_validation(self, is_valid, attr_name):
        """Receive validation error signal from  a panel item and append to validation data to be checked later"""

        self.validation_data.append({"is_valid": is_valid, "msg": attr_name})

    def save_data_in_scenario(self, data, agent_id):
        """Save data in scenario"""

    def recursive_dict_iter(self):
        """Iterate over nested attributes in a dict"""
        for key, value in self.items():
            if isinstance(value, dict):
                yield from self.recursive_dict_iter(value)
            else:
                yield key, value

    def merge_data(self, data):
        """Merge nested attributes in a dict"""
        merged_list = []

        for item in data:
            for key, value in item.items():
                if isinstance(value, list):
                    merged_dict = {}
                    for sub_item in value:
                        merged_dict.update(sub_item)
                    merged_list.append(merged_dict)
                elif isinstance(value, dict):
                    merged_list.extend(self.merge_data([value]))

        return merged_list

    def _group_invalid_data(self):
        """Group invalid data in a string to display in a alert"""
        start_str = "The Following attributes are not valid:\n"

        for item in self.validation_data:
            for item_key, item_value in item.items():
                if item_key != "msg":
                    continue
                if not item["is_valid"]:
                    start_str += "- {}\n".format(item["msg"])

        return start_str

    def on_ok_clicked(self):
        """Processing User input and save data in scenario"""

        self.validation_data.clear()

        self.trigger_input_validation.emit()

        for validated_item in self.validation_data:
            for item_key, item_value in validated_item.items():
                if isinstance(item_value, bool):
                    if not item_value:
                        gen_quick_warning_modal(
                            self, "Validation error", self._group_invalid_data()
                        )

                        return

        self.accepted.emit()

        self._ui.accept()

    def adjust_tree_size(self):
        """Initially  Adjust treeWidget height to fit items"""

        total_height = 0
        iterator = QtWidgets.QTreeWidgetItemIterator(self._ui.treeWidget)
        while iterator.value():
            item = iterator.value()
            row = self._ui.treeWidget.indexOfTopLevelItem(item)
            total_height += self._ui.treeWidget.sizeHintForRow(row)
            iterator += 1

        # Adjust for header height and any additional margins or spacing
        total_height += self._ui.treeWidget.header().height()
        total_height += (
                self._ui.treeWidget.contentsMargins().top()
                + self._ui.treeWidget.contentsMargins().bottom()
        )

        # Set the height of the QTreeWidget
        self._ui.treeWidget.setFixedHeight(total_height)
        self.fired_once = True

    def _reset_attributes(self):
        """Reset UI to default state and rebuild tree view"""

        if self.trigger_input_validation is not None:

            self.trigger_input_validation.disconnect()

        self._ui.treeWidget.clear()
        self._tree_items.clear()

        current_agent_type = self._ui.comboBoxType.currentText()

        for attr_name, attr_spec in self._schema.agent_types[
            current_agent_type
        ].attributes.items():
            agent_type = self._schema.agent_type_from_name(current_agent_type)

            sub_dict = get_sub_attr(agent_type.attributes.items(), None, current_agent_type)

            item = AttributeTreeItem(
                self._ui.treeWidget,
                attr_name,
                attr_spec,
                self.scenario,
                self._working_dir,
                validation_signal=self.trigger_input_validation,
                validation_error_signal=self.validation_error_signal,
                flat_default_values=sub_dict[attr_name],

            )
            self._tree_items.append(item)

        self._ui.treeWidget: QtWidgets.QTreeWidget

        self._ui.treeWidget.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )

    def update(self) -> None:
        """Update tree view and adjust size"""

        self.adjust_tree_size()
        super().update()

    def make_new_agent(self, agent_id) -> models.Agent:
        """Create a new agent with the given ID if it has a value. Note also used for editing agents"""
        agent_type = self._ui.comboBoxType.currentText()
        agent = models.Agent(agent_id, agent_type)

        for item in self._tree_items:
            if item.attr_value is not None:

                if isinstance(item.attr_value, dict):
                    if not item.attr_value:
                        continue

                attr = fameio.Attribute(item.attr_name, item.attr_value)

                agent.add_attribute(item.attr_name, attr)
        return agent



import fameio.input.scenario as fameio


class Contract(fameio.Contract):
    """Extends fameio.Contract with the features required for the GUI"""

    @classmethod
    def from_dict(cls, definitions: dict) -> "Contract":
        return super().from_dict(definitions)

    FIELD_TYPE_INT = "int"
    FIELD_TYPE_STRING = "string"
    FIELD_TYPE_STRING_COMBO_SELECTION = "stringComboSelection"

    FIELD_TYPE_FLOAT = "float"
    FIELD_TYPE_DATE_TIME = "datetime"

    def get_field_types(self) -> list:
        """Returns a list of tuples with the field name and the field type to build a UI"""
        return [
            (super().KEY_SENDER, self.FIELD_TYPE_INT),
            (super().KEY_RECEIVER, self.FIELD_TYPE_INT),
            (super().KEY_PRODUCT, self.FIELD_TYPE_STRING_COMBO_SELECTION),
            (super().KEY_FIRST_DELIVERY, self.FIELD_TYPE_DATE_TIME),
            (super().KEY_INTERVAL, self.FIELD_TYPE_INT),
            (super().KEY_EXPIRE, self.FIELD_TYPE_DATE_TIME),
        ]

    def get_field_type(self, field_name):
        """Returns field type"""
        for field in self.get_field_types():
            if field[0] == field_name:
                return field[1]

        raise RuntimeError(f"Field type not found with the field_name: {field_name}")


    def get_fields_to_edit(self):
        return [
            super().KEY_SENDER,
            super().KEY_RECEIVER,
            super().KEY_PRODUCT,
            super().KEY_FIRST_DELIVERY,
            super().KEY_INTERVAL,
            super().KEY_EXPIRE,
        ]



import logging
import math
import os
import typing
from copy import deepcopy

import fameio.input.validator as fameio
from PySide6 import QtCore
from PySide6.QtWidgets import QApplication
from fameio.input.schema import AttributeSpecs

from famegui import models
from famegui.agent_controller import AgentController
from famegui.appworkingdir import AppWorkingDir
from famegui.config.runtime_consts import UPDATED_CONTRACT_KEY, OLD_CONTRACT_KEY
from famegui.database.prebuild_queries.agent_colors_queries import colors_for_agent_types_exit_in_db, \
    bulk_insert_colors, get_all_colors_for_agent_types, get_color_for_agent_type
from famegui.manager.UndoManager import bulk_save_old_agent_coords, UndoManager
from famegui.models import Agent, Contract, GeneralProperties
from famegui.undo_utils.contract_utils import swap_or_delete_contract


class ProjectProperties(QtCore.QObject):
    """Class used to attach extra properties to a scenario model and signal when they change"""

    changed = QtCore.Signal()

    def __init__(self):
        super().__init__()
        self.reset()

    def reset(self, file_path=""):
        self._has_unsaved_changes = False
        self._file_path = file_path
        self._validation_errors = []
        self.changed.emit()

    @property
    def has_unsaved_changes(self) -> bool:
        return self._has_unsaved_changes

    def set_unsaved_changes(self, has_changes: bool = True):
        if self._has_unsaved_changes != has_changes:
            self._has_unsaved_changes = has_changes
            self.changed.emit()

    @property
    def file_path(self) -> str:
        return self._file_path

    def set_file_path(self, file_path: str) -> None:
        if self._file_path != file_path:
            self._file_path = file_path
            self.changed.emit()

    @property
    def is_validation_successful(self) -> bool:
        return len(self._validation_errors) == 0

    @property
    def validation_errors(self) -> typing.List[str]:
        return self._validation_errors

    def clear_validation_errors(self) -> None:
        was_invalid = not self.is_validation_successful
        self._validation_errors = []
        if was_invalid:
            logging.info("schema validation status changed to valid")
            self.changed.emit()

    def set_validation_error(self, msg: str) -> None:
        assert msg != ""
        was_valid = self.is_validation_successful
        self._validation_errors = [msg]
        if was_valid:
            logging.info("schema validation status changed to invalid")
            self.changed.emit()


class MainController(QtCore.QObject):
    selected_agent_changed = QtCore.Signal(AgentController)
    agent_added = QtCore.Signal(AgentController)
    contract_added = QtCore.Signal(
        AgentController, AgentController, models.Contract
    )  # sender, receiver, contract
    layout_rearranged = QtCore.Signal()
    agents_deleted = QtCore.Signal(list)
    contract_edited = QtCore.Signal(dict)
    contracts_deleted = QtCore.Signal(int)  # agent_id

    class AgentTypeBucket:
        """Class used to store information about a type of agent in the scenario.
        Used to build up orbitals in the graph view
        """

        def __init__(self, type_name: str, agent: Agent = None, agent_list=None):
            self.type_name = type_name
            self.agent_list = [agent]
            if agent_list is not None:
                self.agent_list = agent_list

        type_name: str
        degree_per_agent: float
        radius_scale_factor_per_agent_type: float
        agent_list: typing.List[Agent]
        contract_added = QtCore.Signal(
            AgentController, AgentController, models.Contract
        )  # sender, receiver, contract
        layout_rearranged = QtCore.Signal()
        agents_deleted = QtCore.Signal(list)
        contracts_deleted = QtCore.Signal(int)  # agent_id

        class AgentTypeBucket:
            """Class used to store information about a type of agent in the scenario.
            Used to build up orbitals in the graph view
            """

            def __init__(self, type_name: str, agent: Agent = None, agent_list=None):
                self.type_name = type_name
                self.agent_list = [agent]
                if agent_list is not None:
                    self.agent_list = agent_list

            type_name: str
            degree_per_agent: float
            radius_scale_factor_per_agent_type: float
            agent_list: typing.List[Agent]

    def __init__(self, working_dir: AppWorkingDir):
        super().__init__()
        logging.debug("initializing main controller")
        self._working_dir = working_dir
        os.chdir(working_dir.root_dir)
        self._scenario_model = None

        self._agents: typing.Dict[int, AgentController] = {}
        self._contracts = {}
        self._project_properties = ProjectProperties()
        self._undo_manager = UndoManager(self)

    def reset(self, scenario: models.Scenario = None) -> None:
        self._agents = {}
        self._contracts = {}
        self._scenario_model = scenario
        self._project_properties.reset()

    @property
    def is_open(self) -> bool:
        return self._scenario_model is not None

    @property
    def has_unsaved_changes(self) -> bool:
        return self.project_properties.has_unsaved_changes

    @property
    def project_properties(self) -> ProjectProperties:
        return self._project_properties

    @property
    def schema(self) -> models.Schema:
        return self.scenario.schema

    @property
    def scenario(self) -> models.Scenario:
        return self._scenario_model

    def update_scenario_properties(self, props: models.GeneralProperties):
        has_changed = self._scenario_model.update_properties(props)
        if has_changed:
            self._project_properties.set_unsaved_changes(True)

    @property
    def can_export_protobuf(self) -> bool:
        return (
                self.is_open
                and self.project_properties.is_validation_successful
                and not self.has_unsaved_changes
        )

    @property
    def agent_count(self) -> int:
        """get total agent count of scenario"""
        return len(self._agents)

    @property
    def agent_list(self) -> typing.List[AgentController]:
        return self._agents.values()

    def agent_from_id(self, id: int) -> AgentController:
        """get agent by id"""
        assert id in self._agents
        return self._agents[id]

    @property
    def contract_count(self) -> int:
        """get total contract count of scenario"""
        return len(self._contracts)

    def get_agent_ctrl(self, agent_id: int) -> AgentController:
        """get agent controller by agent id"""
        return self._agents[agent_id]

    def get_agent_type(self, agent_id: int) -> str:
        """get agent type by agent id"""
        return self._agents[agent_id].type_name

    def check_if_all_agents_have_same_type(self, agent_id_list: list) -> bool:
        """check if all agents have the same type"""
        agent_type = self.get_agent_type(agent_id_list[0])
        for agent_id in agent_id_list:
            if self.get_agent_type(agent_id) != agent_type:
                return False
        return True

    def perform_contract_fields_undo(self, current_contract: dict, original_contract: dict):
        """ Performs an undo OP on the requested contract on
        the representative and storage/scenario level"""

        contract_update = {
            UPDATED_CONTRACT_KEY: original_contract,
            OLD_CONTRACT_KEY: current_contract
        }
        self.contract_edited.emit(contract_update)

        swap_or_delete_contract(
            self.scenario,
            current_contract,
            original_contract

        )

    def perform_agent_attributes_undo(self, data_dict: dict):
        self.scenario.update_agent_from_dict(data_dict)

    def perform_redo(self):
        for agent_id, agent_ctrl in self._agents.items():
            agent_ctrl.blockSignals(True)
        self._undo_manager.perform_redo()
        for agent_id, agent_ctrl in self._agents.items():
            agent_ctrl.blockSignals(False)

    def perform_undo(self):

        for agent_id, agent_ctrl in self._agents.items():
            agent_ctrl.blockSignals(True)

        self._undo_manager.perform_undo()

        self.set_unsaved_changes(True)

        for agent_id, agent_ctrl in self._agents.items():
            agent_ctrl.blockSignals(False)

    def get_all_agents_by_type_name(self, agent_type_name: str) -> typing.List[Agent]:
        """filter to get all agents from scenario by agent type"""
        is_agent_from_agent_type = lambda agent: agent.type_name == agent_type_name
        return list(filter(is_agent_from_agent_type, self.scenario.agents))

    def _get_agent_bucket_list(self, agent_list: list) -> list:
        """create agent buckets by agent type"""
        agent_type_bucket_list = []

        for agent in agent_list:
            for x in agent_type_bucket_list:
                if x.type_name == agent.type_name:
                    x.agent_list.append(agent)
                    break
            else:
                agent_type_bucket_list.append(
                    self.AgentTypeBucket(agent.type_name, agent)
                )
        return agent_type_bucket_list

    def prepare_and_sort_agent_buckets(self, agent_type_bucket_list):
        """sort agent buckets by length of agent list"""
        new_agents_list = []
        for item in agent_type_bucket_list:
            new_agents_list.append(item.agent_list)
        new_agents_list.sort(key=len)
        return new_agents_list

    @staticmethod
    def _split_agent_types_into_chunks(agent_type_bucket_list):
        """split huge agent bucket into separate to make canvas look better
        create more buckets to make more smaller orbitals"""

        agent_types_to_re_gen_list = []
        chunked_list = []

        for item in agent_type_bucket_list:
            if len(item.agent_list) > 360:
                agent_types_to_re_gen_list.append(item.type_name)

                for i in range(0, len(item.agent_list), 360):
                    chunked_list.append(item.agent_list[i: i + 360])
        return agent_types_to_re_gen_list, chunked_list

    @staticmethod
    def _calc_degree_per_agent_in_agent_circle(agent_type_bucket_list):
        """calculate degree per agent in agent orbitale"""
        for item in agent_type_bucket_list:
            item.degree_per_agent = 360 / len(item.agent_list)

    @staticmethod
    def _update_graph_items(agent_ctrl_list, new_agent_x, new_agent_y, agent):
        """Updates the graph item coords of the agent with the supplied id"""
        for agent_ctrl in agent_ctrl_list:
            if agent.id.__eq__(agent_ctrl.id):
                agent_ctrl.scene_item.setPos(new_agent_x, new_agent_y)
                agent_ctrl.model_was_modified.emit()
                agent_ctrl.position_changed()
                break

    def get_amount_of_related_contract(self, agent_id: int):
        """Returns the amount of contracts related to the supplied agent"""
        amount_of_related_contract = 0
        for contract in self.scenario.contracts:
            if contract.sender_id == agent_id or contract.receiver_id == agent_id:
                amount_of_related_contract += 1
        return amount_of_related_contract

    def get_all_related_contracts(self, agent_id: int):
        """'Returns a list of all contracts related to the supplied agent"""
        related_contract_list = []
        for contract in self.scenario.contracts:
            if contract.sender_id == agent_id or contract.receiver_id == agent_id:
                related_contract_list.append(contract)
        return related_contract_list

    def delete_all_related_contracts(self, agent_id: int):
        """clear all related contracts of the supplied agent. emit signal to communicate with canvas"""
        self.contracts_deleted.emit(int(agent_id))
        self.scenario.del_all_related_contracts(int(agent_id))
        self._project_properties.set_unsaved_changes(True)

    def delete_all_related_contracts_from_scenario(self, agent_id: int):
        """clear all related contracts of the supplied agent"""
        self.scenario.del_all_related_contracts(agent_id)

    def remove_contract(self, sender_id: int, receiver_id: int, product_name: str):
        """remove contract from scenario"""
        for contract in self.scenario.contracts:
            if (
                    contract.sender_id == sender_id
                    and contract.receiver_id == receiver_id
                    and contract.product_name == product_name
            ):
                self.scenario.contracts.remove(contract)
            self._project_properties.set_unsaved_changes(True)

    def remove_agent(self, agent_id: int, insert_into_last_step_db=True):
        """remove agent from scenario and delete all related contracts"""

        if insert_into_last_step_db:
            pass

        agent_id = int(agent_id)
        agents_to_delete = []
        self.agents_deleted.emit([agent_id])
        for agent in self.scenario.agents:
            if agent.id == agent_id:
                self.scenario.agents.remove(agent)
                agents_to_delete.append(agent)
                self.delete_all_related_contracts_from_scenario(agent_id)
        self.agents_deleted.emit([agent_id])

        self._project_properties.set_unsaved_changes(True)

    # New sub-function to handle agent positioning
    def _adjust_agent_position(self, item, default_radius, radius, margin, last_x, last_y):
        '''Adjusts the position of the agent in the scene by building a orbital model'''
        for i in range(0, len(item.agent_list)):
            agent = item.agent_list[i]
            degree = item.degree_per_agent * (i + 1)

            new_agent_x, new_agent_y, temp_rad = self._calculate_new_position(
                default_radius, degree, last_x, last_y, radius, margin
            )

            if not self.rearrange_mode:
                agent.set_display_xy(new_agent_x, new_agent_y)
                return new_agent_x, new_agent_y, temp_rad, [agent]

            self._adjust_agent_postions(self._agents, new_agent_x, new_agent_y, agent)

        return last_x, last_y, default_radius, []

    # Helper function to calculate new position
    def _calculate_new_position(self, default_radius, degree, last_x, last_y, radius, margin):
        """Helper function to calculate the new position of the agent"""

        new_agent_x = default_radius * math.cos(math.radians(degree))
        new_agent_y = default_radius * math.sin(math.radians(degree))

        distance = math.sqrt((new_agent_x - last_x) ** 2 + (new_agent_y - last_y) ** 2)
        temp_rad = default_radius

        if last_x != 0.0 and last_y != 0.0:
            while distance <= (radius * 2):
                temp_rad = temp_rad * 1.1
                new_agent_x = temp_rad * math.cos(math.radians(degree))
                new_agent_y = temp_rad * math.sin(math.radians(degree))
                distance = math.sqrt((new_agent_x - last_x) ** 2 + (new_agent_y - last_y) ** 2)

        return new_agent_x, new_agent_y, temp_rad

    def delete_all_related_agents(self, agent_id: int):
        """delete all agents that are directly connected through agents to the passed agent_id"""
        unique_agent_id_set = set()
        related_contract_list = self.get_all_related_contracts(agent_id)
        unique_agent_id_set.add(agent_id)
        for contract in related_contract_list:
            unique_agent_id_set.add(contract.sender_id)
            unique_agent_id_set.add(contract.receiver_id)

        for agent in self.scenario.agents:
            if agent.id in unique_agent_id_set:
                self.scenario.agents.remove(agent)
        self._has_unsaved_changes = True

    def _build_agent_type_bucket_list(self, agent_list: list, agent_type_bucket_list: list):
        """Builds a list that wraps the agent list into agent type buckets to build a agent orbitals model"""
        for sub_list in agent_list:
            agent_type_bucket_list.append(
                self.AgentTypeBucket(sub_list[0].type_name, agent_list=sub_list)
            )
        return agent_type_bucket_list

    def _prep_buckets_for_rearrange(self):
        """split huge agent bucket into separate to make canvas look better by creating orbitals"""
        agent_list: list = self.agent_list
        agent_type_bucket_list = self._get_agent_bucket_list(agent_list)

        new_list = []
        for item in self.prepare_and_sort_agent_buckets(agent_type_bucket_list):
            for x in item:
                new_list.append(x)
        agent_types_to_re_gen_list, chunked_list = self._split_agent_types_into_chunks(
            agent_type_bucket_list
        )

        for type_to_rm in agent_types_to_re_gen_list:
            agent_type_bucket_list = [
                value
                for value in agent_type_bucket_list
                if value.type_name != type_to_rm
            ]

        agent_type_bucket_list = self._build_agent_type_bucket_list(chunked_list, agent_type_bucket_list)

        self._calc_degree_per_agent_in_agent_circle(agent_type_bucket_list)

        return sorted(agent_type_bucket_list, key=lambda x: len(x.agent_list))

    def auto_scale_scenario_graph(
            self, agent_ctrl_list: [AgentController], rearrange_mode=False
    ):
        """initiate auto scale of scenario graph into orbital model"""
        try:
            # predefined zoom steps
            default_radius = 100.0
            radius = 50
            margin = 20
            last_x = 0.0
            last_y = 0.0
            min_distance = (float(radius) * 2) + float(margin)

            prep_buckets_for_rearrange = self._prep_buckets_for_rearrange()
            last_list = []

            for item in prep_buckets_for_rearrange:
                # expand rad for new circle
                new_radius = default_radius * 1.1
                if new_radius - default_radius < min_distance:
                    new_radius = new_radius + radius * 2 + margin
                default_radius = new_radius

                for i in range(0, len(item.agent_list)):
                    agent = item.agent_list[i]
                    degree = item.degree_per_agent
                    for counter in range(0, i):
                        degree = degree + item.degree_per_agent
                    new_agent_x = default_radius * math.cos(math.radians(degree))
                    new_agent_y = default_radius * math.sin(math.radians(degree))
                    distance = math.sqrt(
                        (new_agent_x - last_x) ** 2 + (new_agent_y - last_y) ** 2
                    )
                    temp_rad: float = default_radius

                    if last_x != 0.0 and last_y != 0.0:
                        # adjust till all spacing requirements are met
                        while distance <= (radius * 2):
                            temp_rad = temp_rad * 1.1
                            new_agent_x = temp_rad * math.cos(math.radians(degree))
                            new_agent_y = temp_rad * math.sin(math.radians(degree))
                            distance = math.sqrt(
                                (new_agent_x - last_x) ** 2
                                + (new_agent_y - last_y) ** 2
                            )

                    default_radius = temp_rad
                    last_x = new_agent_x
                    last_y = new_agent_y

                    if not rearrange_mode:
                        agent.set_display_xy(int(new_agent_x), int(new_agent_y))
                        last_list.append(agent)
                        continue

                    self._update_graph_items(
                        agent_ctrl_list, new_agent_x, new_agent_y, agent
                    )

            if not rearrange_mode:
                for a in last_list:
                    self._create_agent_controller(a)
                for c in self._scenario_model.contracts:
                    self._create_contract_model(c)

            self._project_properties.set_unsaved_changes(True)


        except Exception as e:
            logging.error("failed to init the given scenario: {}".format(e))
            self.reset()
            raise

        # refresh the UI
        self._project_properties.changed.emit()

    def compute_scene_rect(self) -> QtCore.QRectF:
        rect = QtCore.QRectF(0, 0, 1000, 1000)

        if len(self.agent_list) >= 5000:
            rect = QtCore.QRectF(0, 0, 100000, 100000)

        for a in self.agent_list:
            margin = 20
            item_size = a.scene_item.boundingRect().width()
            left = a.scene_item.x() - margin
            right = a.scene_item.x() + item_size + margin
            top = a.scene_item.y() + -margin
            bottom = a.scene_item.y() + item_size + margin
            if left < rect.left():
                rect.setLeft(left)
            if right > rect.right():
                rect.setRight(right)
            if top < rect.top():
                rect.setTop(top)
            if bottom > rect.bottom():
                rect.setBottom(bottom)

        return rect

    def get_contract(self, sender_id, receiver_id, product_name) -> Contract:
        for i in self.scenario.contracts:
            i: Contract
            if i.product_name.__eq__(product_name):
                if i.sender_id.__eq__(sender_id) and i.receiver_id.__eq__(receiver_id):
                    return i

                elif i.sender_id.__eq__(receiver_id) and i.receiver_id.__eq__(
                        sender_id
                ):
                    return i

    def get_help_text_for_attr(self, agent_type_name: str, attr_name: str) -> str:
        """Extracts the help text for the given attribute name of the given agent type name"""
        for name, spec in self.schema.agent_type_from_name(
                agent_type_name
        ).attributes.items():
            spec: AttributeSpecs
            if name.__eq__(attr_name):
                return spec.help_text

    # self.schema.agent_types

    def generate_new_agent_id(self):
        new_id = len(self._agents) + 1
        # note: we don't control how ids have been generated for agents created from an external source
        # so we check for possible conflict and solve it
        if new_id in self._agents:
            for i in range(1, len(self._agents) + 2):
                if i not in self._agents:
                    new_id = i
                    break
        logging.debug("generated new agent id {}".format(new_id))
        return new_id

    # the given agent id can be 0 to clear the current selection
    def set_selected_agent_id(self, agent_id: int):

        assert self.is_open
        if agent_id not in self._agents:
            assert agent_id == 0 or agent_id is None
            self.selected_agent_changed.emit(None)
        else:
            self.selected_agent_changed.emit(self._agents[agent_id])

    def add_new_agent(self, agent_model: models.Agent, x: int, y: int):
        assert self.is_open
        agent_model.set_display_xy(x, y)

        color_for_type = get_color_for_agent_type(agent_model.type_name)

        self._create_agent_controller(agent_model, color_for_type)

        self._scenario_model.add_agent(agent_model)
        self._project_properties.set_unsaved_changes(True)
        self.revalidate_scenario()

        logging.info(
            "created new agent {} of type '{}'".format(
                agent_model.display_id, agent_model.type_name
            )
        )

    def set_agent_display_xy(self, agent_id: int, x: float, y: float):

        self._adjust_agent_postions(self.agent_list, x, y, self.get_agent_ctrl(agent_id).model)

    def agent_model_was_modified(self, agent_id: int, x=-1, y=-1):

        ctrl = self.get_agent_ctrl(agent_id)

        if x != -1:
            if not QApplication.instance().mouseButtons() & QtCore.Qt.MouseButton.LeftButton:
                agent = self.scenario.get_agent_by_id(agent_id)

                old_x = agent.display_xy[0]
                old_y = agent.display_xy[1]
                self.project_properties.set_unsaved_changes(True)
                self._undo_manager.save_moved_agent(x, y, agent_id, old_x, old_y)

    def save_project_properties_manipulation(self, original_dict: dict, updated_dict: dict):

        self._undo_manager.save_project_properties_manipulation(
            original_dict,
            updated_dict
        )

    def _create_agent_controller(self, agent_model: models.Agent,
                                 svg_color: typing.Union[str, None] = None):
        assert self.is_open

        # accept to add the agent even if invalid
        agent_ctrl = AgentController(agent_model, svg_color)
        self._agents[agent_ctrl.id] = agent_ctrl

        agent_ctrl.model_was_modified.connect(lambda: self.agent_model_was_modified(agent_ctrl.id))
        agent_ctrl.model_was_moved.connect(lambda x, y: self.agent_model_was_modified(agent_ctrl.id, x, y))

        logging.info("new agent {} added".format(agent_ctrl.display_id))

        self.agent_added.emit(agent_ctrl)

    def add_new_contract(self, contract_model: models.Contract):
        self._scenario_model.add_contract(contract_model)
        self._create_contract_model(contract_model)
        self._project_properties.set_unsaved_changes(True)
        self.revalidate_scenario()
        logging.info(
            "created new contract '{}' between {} and {}".format(
                contract_model.product_name,
                contract_model.display_sender_id,
                contract_model.display_receiver_id,
            )
        )

    def set_unsaved_changes(self, unsaved: bool):
        """set the unsaved changes flag for the current scenario"""
        self._project_properties.set_unsaved_changes(unsaved)

    def _create_contract_model(self, contract: models.Contract):
        assert self.is_open

        # validate sender / receiver are known
        if contract.sender_id not in self._agents:
            raise ValueError(
                "can't add contract '{}' because sender agent id '{}' is unknown".format(
                    contract.product_name, contract.sender_id
                )
            )

        if contract.receiver_id not in self._agents:
            raise ValueError(
                "can't add contract '{}' because receiver agent id '{}' is unknown".format(
                    contract.product_name, contract.receiver_id
                )
            )

        sender_ctrl = self._agents[contract.sender_id]
        receiver_ctrl = self._agents[contract.receiver_id]

        # connect agents
        sender_ctrl.model.add_output(contract.receiver_id)
        receiver_ctrl.model.add_input(contract.sender_id)

        self.contract_added.emit(sender_ctrl, receiver_ctrl, contract)

    def apply_properties(self, props_to_apply: dict):

        general_props = GeneralProperties.from_dict(props_to_apply)

        self.scenario.update_properties(general_props)

    def re_render_scene(self, new_coords_list):
        """Re-render scene with parsed agents"""

        agent_list: list = list(self.agent_list)
        for idx, item in self._agents.items():
            x = new_coords_list[idx]["agent_x"] if idx in new_coords_list else new_coords_list[str(idx)]["agent_x"]
            y = new_coords_list[idx]["agent_y"] if idx in new_coords_list else new_coords_list[str(idx)]["agent_y"]

            self._adjust_agent_postions(agent_list, x, y, item.model)

    def rearrange_layout(self):
        """rearrange agents in the graph view into the orbital layout"""

        copied_agents = self.agent_list

        for item in copied_agents:
            item.blockSignals(True)

        last_cords_list = []

        for item in copied_agents:
            last_cords_list.append({
                "agent_id": deepcopy(item.id),
                "agent_x": deepcopy(item.scene_item.x()),
                "agent_y": deepcopy(item.scene_item.y()),
            })

        agent_ctrl_dict = {item["agent_id"]: {
            "agent_x": deepcopy(item["agent_x"]),
            "agent_y": deepcopy(item["agent_y"]),
        } for item in last_cords_list}

        self.auto_scale_scenario_graph(self.agent_list, True)

        for item in copied_agents:
            item.blockSignals(False)
            item.model_was_modified.connect(lambda: self.agent_model_was_modified(item.id))
            item.model_was_moved.connect(lambda x, y: self.agent_model_was_modified(item.id, x, y))

        for item in self.agent_list:
            last_cords_list.append({
                "agent_id": deepcopy(item.id),
                "agent_x": deepcopy(item.scene_item.x()),
                "agent_y": deepcopy(item.scene_item.y()),
            })

        agent_ctrl_dict_post_transformational_coords = {item["agent_id"]: {
            "agent_x": deepcopy(item["agent_x"]),
            "agent_y": deepcopy(item["agent_y"]),
        } for item in last_cords_list}

        bulk_save_old_agent_coords(agent_ctrl_dict, agent_ctrl_dict_post_transformational_coords)

    def revalidate_scenario(self) -> bool:
        assert self._scenario_model is not None

        try:
            fameio.SchemaValidator.validate_scenario_and_timeseries(self._scenario_model)
            self._project_properties.clear_validation_errors()
            return True
        except fameio.ValidationError as e:
            err_msg = str(e)
            logging.warning("failed to validate the scenario: {}".format(err_msg))
            self._project_properties.set_validation_error(err_msg)
            return False

    def get_agent_type_to_color_with_type_listing(self) -> tuple[dict, set]:
        pre_defined_agent_colors = get_all_colors_for_agent_types()

        pre_defined_agent_colors = pre_defined_agent_colors if pre_defined_agent_colors is not None else []

        ui_type_to_color_mapping = {item["agent_type_name"]: item["agent_color_hex_code"] for item in
                                    pre_defined_agent_colors}

        agent_type_names_from_db = set([
            item["agent_type_name"] for item in pre_defined_agent_colors
        ])

        return ui_type_to_color_mapping, agent_type_names_from_db

    def get_agent_ids_to_type(self) -> dict:

        return {agent_ctrl.id: agent_ctrl.type_name for agent_ctrl in self.agent_list}

    def _init_scenario_model_with_non_auto_arrange(self):
        """initialize the scenario model from the current state of the scenario coords
        loads agent controller initially
        """

        populate_colors = False

        ui_type_to_color_mapping, agent_type_names_from_db = self.get_agent_type_to_color_with_type_listing()

        if not colors_for_agent_types_exit_in_db():
            populate_colors = True

        for a in self._scenario_model.agents:
            svg_color = ui_type_to_color_mapping[a.type_name] if a.type_name in ui_type_to_color_mapping else None
            self._create_agent_controller(a, svg_color)

        type_to_color_dict = {item.type_name: item.svg_color for key, item in self._agents.items() if
                              not item.type_name in agent_type_names_from_db}

        new_agent_type_to_color_dict = {agent_ctrl.model.type_name: agent_ctrl.svg_color for agent_ctrl in
                                        self.agent_list if
                                        agent_ctrl.model.type_name not in agent_type_names_from_db}

        if populate_colors:
            bulk_insert_colors(type_to_color_dict)

        if new_agent_type_to_color_dict and not populate_colors:
            bulk_insert_colors(new_agent_type_to_color_dict)

        # bulk insert generated colors

        for c in self._scenario_model.contracts:
            self._create_contract_model(c)

        return

    def _adjust_agent_postions(
            self,
            agent_ctrl_list: typing.List[AgentController],
            new_agent_x: float,
            new_agent_y: float,
            agent,
    ):
        """adjust the position of the agent in the scene"""

        for agent_ctrl in agent_ctrl_list:
            if agent.id.__eq__(agent_ctrl.id):
                agent_ctrl.scene_item.setPos(new_agent_x, new_agent_y)
                agent_ctrl.model.set_display_xy(new_agent_x, new_agent_y)

                agent.set_display_xy(new_agent_x, new_agent_y)

                agent_ctrl.position_changed()

                agent_ctrl.scene_item.adjust_contracts()
                agent_ctrl.scene_item.update()

                break

    def init_scenario_model(self, scenario: models.Scenario, file_path: str):
        """init the scenario model and create the corresponding controllers.
        Build the graph orbital model and validate the scenario.
        """

        logging.debug("opening new scenario")
        rearrange_mode = False

        if self.is_open:
            rearrange_mode = True

        self.reset()

        try:
            self._scenario_model = scenario
            self._project_properties.reset(file_path)
            if not rearrange_mode:
                self._init_scenario_model_with_non_auto_arrange()

                return

            try:
                last_list = []
                agent_type_bucket_list = self._prep_buckets_for_rearrange()

                default_radius = 100.0
                radius = 50
                margin = 20
                last_x = 0.0
                last_y = 0.0

                min_distance = (float(radius) * 2) + float(margin)

                for item in agent_type_bucket_list:
                    new_radius = default_radius * 1.1

                    if new_radius - default_radius < min_distance:
                        new_radius = new_radius + radius * 2 + margin
                    default_radius = new_radius

                    new_radius = default_radius * 1.1
                    if new_radius - default_radius < min_distance:
                        new_radius = new_radius + radius * 2 + margin
                    default_radius = new_radius

                    last_x, last_y, default_radius, new_last_list = self._adjust_agent_position(
                        item, default_radius, radius, margin, last_x, last_y
                    )
                    last_list.extend(new_last_list)

                if not rearrange_mode:
                    for a in last_list:
                        self._create_agent_controller(a)
                    for c in self._scenario_model.contracts:
                        self._create_contract_model(c)

            except Exception as e:
                logging.error("failed to init the given scenario: {}".format(e))
                self.reset()
                raise Exception(e)

            # refresh the UI
            self._project_properties.changed.emit()

        except Exception as e:
            logging.error("failed to init the given scenario: {}".format(e))
            self.reset()
            raise

        # refresh the UI

        self._project_properties.changed.emit()

        self.revalidate_scenario()

    def save_to_file(self, file_path):
        assert self.is_open
        models.ScenarioLoader.save_to_yaml_file(self._scenario_model, file_path)
        # update status
        self._project_properties.set_unsaved_changes(False)
        self._project_properties.set_file_path(file_path)
        self.revalidate_scenario()




import logging
import re
import typing

import fameio.input.schema as schema
from PySide6 import QtWidgets, QtGui, QtCore
from PySide6.QtCore import QSize, Signal, Qt
from PySide6.QtWidgets import QVBoxLayout, QPushButton, QLabel, QSizePolicy
from fameio.input.scenario import Attribute
from fameio.input.schema import AttributeSpecs

from famegui.appworkingdir import AppWorkingDir
from famegui.config.static_ui_descriptions import NO_SUPPORTED_EDIT_WIDGET_TEXT
from famegui.config.style_config import get_color_for_type, get_border_for_type, ERROR_TEXT_STYLE_LABEL, \
    ERROR_TEXT_STYLE_LINE_EDIT
from famegui.data_manager.string_set_manager import get_pre_defined_string_set_values_from_scenario
from famegui.database.db_loader.db_manager import get_string_set_values_for_group
from famegui.fame_tools.dict_tools import remove_none_values_and_empty_elements
from famegui.fame_tools.string_utils import get_last_substring
from famegui.famegui_widgets.complex_attribute_schema_widget import SchemaWidget
from famegui.ui.block_item_wrapper import BlockItemWrapper
from famegui.ui.block_type_wrapper import BlockTypeWrapper
from famegui.ui.fame_input_panels import FileChooserWidgetPanel
from famegui.ui.fame_ui_elements import QFameBoldLabel, DescriptionLabel
from famegui.ui.list_block import CustomListWidget
from famegui.ui.ui_input import set_placeholder_text
from famegui.ui.ui_input_panel_wrapper import InputPanelWrapper, MiniAttributeSpecs
from famegui_runtime_helper.attribute_dict_formatting_helper import StringAttributeType


class RootAttributeInputPanelWrapper(QtWidgets.QWidget):
    """Wrapper for the root level attribute input panels. It is used to toggle the visibility
    of the input panels and mange all nested input panels.
    """

    def __init__(self, row, panel, parent=None):
        super(RootAttributeInputPanelWrapper, self).__init__(parent)
        self.setAttribute(Qt.WA_StyledBackground, True)
        layout = QVBoxLayout(self)
        self.layout = layout
        self.layout.setSpacing(10)
        self.content_frame: QtWidgets.QWidget = QtWidgets.QWidget(self)

        self.row: AttributeTreeItem = row
        button = QPushButton("Collapse/Expand")
        layout.addWidget(button)
        button.clicked.connect(self.toggle_content)
        layout.addWidget(self.content_frame)

        content_layout = QVBoxLayout(self.content_frame)

        self.panel = panel

        content_layout.addWidget(self.panel)

        self.initial_width = self.width()

        self.content_frame.setVisible(True)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def toggle_size_adjustment(self):
        """Toggle the size adjustment of the parent widget."""
        self.adjustSize()

        self.updateGeometry()
        self.row.setSizeHint(0, QSize(-1, 60))
        self.row.setSizeHint(1, QSize(-1, 60))
        self.parent().setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.parent().updateGeometry()
        self.parent().adjustSize()

    def toggle_content(self):
        """Toggle the visibility of the input panel by triggering layout updates."""

        if self.content_frame.isVisible():
            self.content_frame.hide()

        else:
            self.content_frame.show()

        self.parent().updateGeometry()
        self.parent().adjustSize()
        self.parent().update()

    def height(self):
        """Get the pre-calculated height of the input panel."""
        return self.content_frame.sizeHint().height() + 50


class AttributeTreeItem(QtWidgets.QTreeWidgetItem):
    """Root Holder class for all attribute tree items.
    It is used to create a tree of attributes and their input panels to access all attributes .
    """

    def simple_attr_build(self, attr_spec: schema.AttributeSpecs, result_dict):

        if attr_spec.attr_type == schema.AttributeType.BLOCK:
            result_dict_sub = {}

    def set_schema_widget_height(self, height: int):

        if self.is_single_item:
            self.ui_parent.itemWidget(self, 1).setFixedHeight(int(height))
            self.ui_parent.itemWidget(self, 1).update()
            self.ui_parent.itemWidget(self, 1).updateGeometry()
            height = height - 90
            self._schema_widget.setFixedHeight(int(height))



    def __init__(
            self,
            parent: QtWidgets.QTreeWidget,
            attr_name: str,
            attr_spec: schema.AttributeSpecs,
            scenario,  # Scenario
            working_dir: AppWorkingDir,
            onInputChanged: Signal = None,
            validation_signal: Signal = None,
            validation_error_signal: Signal = None,
            default_value=None,
            flat_default_values: dict = None,
            is_single_item: bool = False
    ):
        self.state_value_dict = {}
        self._attr_name = attr_name
        self._attr_spec: AttributeSpecs = attr_spec
        self.validation_signal: Signal = validation_signal
        self.validation_error_signal: Signal = validation_error_signal
        self._working_dir = working_dir
        self._default_value = default_value
        self._attr_value = None
        self._display_error = lambda has_error: None
        self.initial_height = 0
        self.schema = scenario.schema
        self.scenario = scenario
        self.holder_widget = None
        self.default_agent_flat_value_dict = flat_default_values

        self.is_single_item = is_single_item

        self.onInputChanged = onInputChanged

        QtWidgets.QTreeWidgetItem.__init__(self, parent, [attr_name])

        font = self.font(0)
        if self._attr_spec.is_mandatory:
            font.setBold(True)

            tooltip = self.tr("{} (mandatory)").format(self._attr_name)
        else:
            font.setItalic(True)
            tooltip = self.tr("{} (optional)").format(self._attr_name)

        self.setFont(0, font)
        self.setToolTip(0, tooltip)
        self.ui_parent = parent
        self.panel = QLabel("Hello")  # ForeCastAPi -> return None

        self.panel = QLabel()

        self.holder_widget = QtWidgets.QWidget()

        layout = QVBoxLayout()
        placeholder = QtWidgets.QWidget()
        placeholder.setStyleSheet("background-color: #0A0A0A;")
        placeholder.setContentsMargins(0, 0, 0, 0)  # Adds padding inside the placeholder layout
        inner_layout = QVBoxLayout()

        placeholder.setLayout(inner_layout)

        layout.addWidget(placeholder, stretch=1)

        self.holder_widget.setLayout(layout)

        color_str = get_color_for_type(self._attr_spec)

        self.holder_widget.setStyleSheet(
            f"RootAttributeInputPanelWrapper {{  background:{color_str}; border-radius: 5px;"
            "margin-top:10px; ; margin-left: 5px; margin-right: 5px; }}"
        )

        # Create a boundary container
        boundary_container = QtWidgets.QWidget()
        boundary_layout = QVBoxLayout(boundary_container)
        boundary_layout.setContentsMargins(0, 0, 0, 0)
        boundary_layout.setSpacing(0)

        # Set maximum height to constrain the widget
        # boundary_container.setMaximumHeight(400)  # Adjust the height as needed

        # Add the SchemaWidget to the boundary container
        self._schema_widget = SchemaWidget(self.default_agent_flat_value_dict,
                                           is_single_item=is_single_item,
                                           attr_row_item=self)

        self._schema_widget.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)

        boundary_layout.addWidget(self._schema_widget)
        boundary_layout.addStretch(1)
        parent.header().setSectionResizeMode(QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        parent.header().setStretchLastSection(True)

        example_widget = QtWidgets.QWidget()
        example_layout_v = QVBoxLayout()
        example_layout_v.setAlignment(Qt.AlignmentFlag.AlignTop)
        example_layout_v.setContentsMargins(0, 0, 0, 0)  # Remove margins
        example_layout_v.setSpacing(0)  # Remove spacing
        # Add stretch to fill space dynamically

        example_widget.setLayout(example_layout_v)

        example_widget.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        boundary_container.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)

        # Set the boundary container as the item widget
        example_layout_v.addWidget(boundary_container, stretch=1)

        parent.setItemWidget(self, 1, example_widget)

        # Set the size hint for the tree widget item to match the boundary container
        # self.setSizeHint(1, boundary_container.sizeHint())

    def get_attr_spec(self):
        """Get the root AttributeSpecs of the row."""
        return self._attr_spec

    @property
    def attr_name(self) -> str:
        """Get the attribute short name of the root attribute ."""
        return self._attr_name

    def remove_none_values(self, d):
        """Helper function to remove None values from a dictionary."""
        if not isinstance(d, dict):
            return d
        return {k: self.remove_none_values(v) for k, v in d.items() if v is not None}



    @staticmethod
    def get_substring_after_last_dot(s):
        if "." in s:
            return s.rsplit(".", 1)[-1]
        return s

    def build_it(self, attr: AttributeSpecs, li_idx=-1):

        res_dict = {}



        if attr.is_list and attr.has_nested_attributes:

            max_inputs = 0
            data_count = 0

            list_data = []

            while data_count <= max_inputs:
                sub_list_item = {}

                for sub_key, sub_item in attr.nested_attributes.items():

                    if sub_item.attr_type == schema.AttributeType.BLOCK:
                        sub_list_item[sub_key] = self.build_it(sub_item, li_idx=data_count)
                        continue

                    sub_full_name = sub_item.full_name

                    value_list = self.state_value_dict[
                        sub_full_name] if sub_full_name in self.state_value_dict else None  # fallback in case optional list nodes where removed

                    if value_list is None:
                        continue

                    if isinstance(value_list, InputPanelWrapper):
                        sub_list_item[sub_key] = value_list.get_input_value()
                        continue

                    if not isinstance(value_list, list):
                        sub_list_item[sub_key] = value_list[data_count].get_input_value()
                        continue

                    amount_of_inputs = len(value_list) if value_list is not None else 0

                    if amount_of_inputs == 0:
                        break
                    if max_inputs == 0:
                        max_inputs = (amount_of_inputs - 1)

                    sub_list_item[sub_key] = value_list[data_count].get_input_value()

                data_count += 1
                list_data.append(sub_list_item)

            attr_name = self.get_substring_after_last_dot(attr.full_name)

            res_dict[attr_name] = list_data

        elif attr.has_nested_attributes:
            for sub_key, sub_item in attr.nested_attributes.items():

                if sub_item.attr_type == schema.AttributeType.BLOCK:
                    res_dict[sub_key] = self.build_it(sub_item)
                    continue

                value = self.state_value_dict[
                    sub_item.full_name] if sub_item.full_name in self.state_value_dict else None

                if value is None:
                    res_dict[sub_key] = None
                    continue

                if sub_item.is_list:
                    if isinstance(value, list):
                        res_dict[sub_key] = [
                            list_iter_item.get_input_value() for list_iter_item in value]
                        continue
                    res_dict[sub_key] = [value.get_input_value()]  # single value case

                    continue

                if isinstance(value, list):
                    if len(value) == 0:
                        res_dict[sub_key] = None
                        continue

                res_dict[sub_key] = value.get_input_value() if not isinstance(value, list) else value[
                    li_idx].get_input_value()

        else:
            res_dict[self.attr_name] = self.state_value_dict[self._attr_spec.full_name].get_input_value()

        return res_dict

    @property
    def attr_value(self) -> typing.Any:
        """Get clean attribute value from a complete row"""

        res_dict = self.build_it(self._attr_spec)

        cleaned_data = remove_none_values_and_empty_elements(res_dict)

        if self._attr_spec.attr_type != schema.AttributeType.BLOCK:
            return cleaned_data[next(iter(cleaned_data.keys()))] if len(cleaned_data.keys()) != 0 else cleaned_data

        cleaned_data = cleaned_data[self.attr_name] if self.attr_name in cleaned_data else cleaned_data

        return cleaned_data

    @property
    def validation_error(self) -> typing.Optional[str]:
        if self._attr_spec.is_mandatory and self._attr_value is None:
            return "mandatory attribute '{}' is missing".format(self.attr_name)
        return None

    _supported_list_types = [
        schema.AttributeType.INTEGER,
        schema.AttributeType.LONG,
        schema.AttributeType.DOUBLE,
        schema.AttributeType.STRING,
        schema.AttributeType.STRING_SET,
    ]

    @property
    def _attr_enum_values(self) -> typing.List[str]:
        """Returns the enum values of the attached AttributeSpecs"""
        assert self._attr_spec.attr_type == schema.AttributeType.ENUM
        assert not self._attr_spec.is_list
        assert not self._attr_spec.has_nested_attributes
        return self._attr_spec.values

    def build_parent_container(self, attr_spec_inner, attr_name: str, holder_widget) -> QtWidgets.QWidget:
        """Build a parent container for a block type attribute with an n depth of nested attributes"""

        value = attr_spec_inner.help_text or ""
        desc = DescriptionLabel(text=value)

        label = QFameBoldLabel(
            text=f"{attr_name}:x: {attr_spec_inner.attr_type.name} "
        )

        label.setWordWrap(True)

        parent_layout = QVBoxLayout()

        parent_container = BlockTypeWrapper(
            block_type_layout=parent_layout,
            spec=attr_spec_inner,
            attr_name=attr_name,
        )

        layout_header = QVBoxLayout()
        header_widget = QtWidgets.QWidget()

        header_widget.setLayout(layout_header)

        layout_header.insertWidget(0, label)
        layout_header.insertWidget(1, desc)
        parent_layout.insertWidget(0, header_widget)

        # Wrap holder_widget with a parent container and chevron icon

        # Set the stylesheet for the specific widget using its object name

        chevron_button = QPushButton(
            "Expand/Collapse"
        )  # Using a triangle as a placeholder for the chevron icon

        chevron_button.setCheckable(True)

        # Set the initial checked state to match the initial visibility of holder_widget
        chevron_button.setChecked(True)

        chevron_button.toggled.connect(holder_widget.setVisible)
        parent_layout.addWidget(chevron_button)
        parent_layout.addWidget(holder_widget)

        return parent_container

    def build_child_sub_block_list(self, attr_spec_inner, attr_name, holder: QVBoxLayout, list_data=None):
        """Build a child list block panel for a block"""

        preset_value_list = self.get_default_value(
            attr_spec_inner, attr_spec_inner.full_name, list_data
        )

        block_list = self.build_sub_block_list(
            attr_spec_inner, preset_value_list
        )

        dynamic_list = CustomListWidget(
            block_list, attr_spec_inner, self.schema,
            attr_name=attr_name
        )

        dynamic_list.row_creation_requested.connect(self._add_list_row)

        holder.addWidget(dynamic_list)

        # dynamic_list.setLayout(holder)

    def build_primitive_block_child(self, attr_spec_inner: AttributeSpecs,
                                    attr_name: str, holder: QVBoxLayout,
                                    list_data=None):
        """Build a primitive child input panel for a block"""
        default_row = QFameBoldLabel(
            text=f"{attr_name}: {attr_spec_inner.attr_type.name}"
        )

        default_row.setWordWrap(True)
        value = attr_spec_inner.help_text or ""

        desc = DescriptionLabel(text=value)

        parent_widget = BlockItemWrapper(spec=attr_spec_inner, attr_name=attr_name)

        parent_widget.add_widget(default_row)
        parent_widget.add_widget(desc)
        prim_widget = self.get_primitive_input_panel(attr_spec_inner, list_data)

        uniq_widget = f"myUniqueWidget-{id(parent_widget)}"

        parent_widget.setObjectName(uniq_widget)

        color_str = get_color_for_type(attr_spec_inner)

        border_style = get_border_for_type(attr_spec_inner)

        parent_widget.setStyleSheet(
            f"#{uniq_widget} {{"
            f"    background-color: {color_str};"
            f"    {border_style}"
            "    border-radius: 5px;"
            "}"
        )

        parent_widget.add_input_widget(prim_widget)

        holder.addWidget(parent_widget)

    def build_sub_block_list(self, attr_data: AttributeSpecs, list_data=None):
        """Build a list of sub blocks for a block type attribute with an n depth of nested attributes"""
        holder = QVBoxLayout()

        if attr_data.has_nested_attributes:
            for item, attr_spec_inner in attr_data.nested_attributes.items():
                attr_spec_inner: AttributeSpecs

                if attr_spec_inner.attr_type == schema.AttributeType.BLOCK:
                    holder_widget = self.build_sub_block_list(attr_spec_inner, list_data)

                    parent_container = self.build_parent_container(
                        attr_spec_inner, item, holder_widget
                    )

                    input_wrapper = InputPanelWrapper(
                        inner_widget=parent_container,
                        signal_to_connect=self.validation_signal,
                        validation_error_signal=self.validation_error_signal,
                        attr_name=item,
                        attribute=attr_spec_inner,
                    )

                    holder.addWidget(input_wrapper)
                    continue

                if attr_spec_inner.is_list:
                    self.build_child_sub_block_list(attr_spec_inner, item, holder, list_data)

                    continue

                self.build_primitive_block_child(attr_spec_inner, item, holder, list_data)

                ## here

        default_data_dict = self.get_default_value(
            attr_data, self._attr_spec.full_name, list_data
        )

        holder_wrapper = BlockTypeWrapper(
            block_type_layout=holder,
            spec=attr_data,
            attr_name=self.attr_name,
            dict_data=default_data_dict,
        )

        return holder_wrapper

    @staticmethod
    def get_substring_from_last_dot(s):
        return s[s.rfind('.') + 1:] if '.' in s else s

    def get_list_capable_int_or_long_input_panel(self, spec: MiniAttributeSpecs,
                                                 nested_list, full_name=None,
                                                 preset_value=None, list_idx=None):

        full_name = full_name if full_name is not None else spec.full_name

        line_edit = self._create_line_edit("0", None)
        line_edit.setText(str(preset_value))
        line_edit.setValidator(QtGui.QIntValidator())

        input_wrapper = InputPanelWrapper(
            inner_widget=line_edit,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
            list_idx=list_idx
        )

        if full_name in self.state_value_dict:
            item = self.state_value_dict[full_name]

            if isinstance(item, list):
                item.append(input_wrapper)
                return line_edit
            new_list = [item]
            new_list.append(input_wrapper)
            self.state_value_dict[full_name] = new_list

            return line_edit

        self.state_value_dict[full_name] = input_wrapper

        return input_wrapper



    def get_list_capable_time_series_input_panel(self, spec: MiniAttributeSpecs, nested_list,
                                                 full_name, preset_value=None, list_idx=None):
        # preset_value = self.get_default_value(spec, spec.full_name, nested_list)

        file_chooser = FileChooserWidgetPanel(
            self._working_dir,
            get_last_substring(full_name),
            preset_value,
            self.ui_parent,
        )

        self._display_error = file_chooser.display_error

        input_wrapper = InputPanelWrapper(
            inner_widget=file_chooser,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
            list_idx=list_idx
        )

        if full_name in self.state_value_dict:
            item = self.state_value_dict[full_name]

            if isinstance(item, list):
                item.append(input_wrapper)
                return file_chooser
            new_list = [item]
            new_list.append(input_wrapper)
            self.state_value_dict[full_name] = new_list

            return file_chooser

        self.state_value_dict[full_name] = input_wrapper

        return file_chooser

    def get_list_capable_enum_input_panel(self, spec: MiniAttributeSpecs,
                                          nested_list, full_name=None,
                                          options=None, preset_value=None, list_idx=None):
        combo_box = QtWidgets.QComboBox()

        # self.schema.get
        full_name, attr_type, is_mandatory, enum_values = spec.get_values()

        combo_box.addItems(enum_values)
        if len(enum_values) == 1:
            combo_box.setCurrentIndex(0)
            self._attr_value = enum_values[0]
        else:
            combo_box.setCurrentIndex(-1)

        preset_value = preset_value if preset_value is not None else None
        set_placeholder_text(combo_box, preset_value, enum_values)

        full_name = full_name if full_name is not None else spec.full_name

        input_wrapper = InputPanelWrapper(
            inner_widget=combo_box,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
            list_idx=list_idx
        )

        if full_name in self.state_value_dict:
            item = self.state_value_dict[full_name]

            if isinstance(item, list):
                item.append(input_wrapper)
                return combo_box
            new_list = [item]
            new_list.append(input_wrapper)
            self.state_value_dict[full_name] = new_list

            return combo_box

        self.state_value_dict[full_name] = input_wrapper

        return combo_box

    def remove_state_registers(self, remove_state_registers,
                               to_remove_idx):

        for full_name in remove_state_registers:
            item = self.state_value_dict[full_name]

            if isinstance(item, list):
                selected = item[to_remove_idx]
                new_list = [sub_item for sub_item in item if sub_item is not selected]
                self.state_value_dict[full_name] = new_list
                continue

            if to_remove_idx != 0:
                raise ValueError("Not possible case")

            del self.state_value_dict[full_name]

    def get_list_capable_double_input_panel(self, spec: MiniAttributeSpecs,
                                            nested_list, preset_value=None, list_idx=None):
        # preset_value = self.get_default_value(spec, spec.full_name, nested_list)
        full_name, attr_type, is_mandatory, _ = spec.get_values()

        line_edit = self._create_line_edit(preset_value, None)

        line_edit.setText(str(preset_value))
        validator = QtGui.QDoubleValidator()
        # accept '.' as decimal separator
        validator.setLocale(QtCore.QLocale.Language.English)
        line_edit.setValidator(validator)

        input_wrapper = InputPanelWrapper(
            inner_widget=line_edit,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
            list_idx=list_idx

        )

        if full_name in self.state_value_dict:
            item = self.state_value_dict[full_name]

            if isinstance(item, list):
                item.append(input_wrapper)
                return line_edit
            new_list = [item]
            new_list.append(input_wrapper)
            self.state_value_dict[full_name] = new_list

            return line_edit

        self.state_value_dict[full_name] = input_wrapper

        return line_edit

    def get_primitive_input_panel(self, spec: typing.Union[AttributeSpecs, None], nested_list,
                                  mini_spec: MiniAttributeSpecs = None, preset_value=None, list_idx=None):

        """Get a primitive input panel for a given attribute spec"""
        full_name, attr_type, is_mandatory, options = mini_spec.get_values()

        preset_value = preset_value if preset_value is not None else ""

        current_attr_type = attr_type
        full_name = full_name

        if current_attr_type == StringAttributeType.ENUM.value:

            # exi 1
            return self.get_list_capable_enum_input_panel(mini_spec, nested_list,
                                                          full_name=full_name, options=options,
                                                          preset_value=preset_value, list_idx=list_idx)

        elif current_attr_type == StringAttributeType.TIME_SERIES.value:

            # exi 2

            resultTS = self.get_list_capable_time_series_input_panel(mini_spec, nested_list,
                                                                     full_name, preset_value, list_idx)

            return resultTS

        elif (current_attr_type == StringAttributeType.INTEGER.value
              or current_attr_type == StringAttributeType.LONG.value):

            # exi 3

            return self.get_list_capable_int_or_long_input_panel(mini_spec, nested_list,
                                                                 full_name, preset_value, list_idx)

        elif current_attr_type == StringAttributeType.DOUBLE.value:

            # exi 4

            return self.get_list_capable_double_input_panel(mini_spec, nested_list, preset_value, list_idx)

        elif current_attr_type == StringAttributeType.STRING.value:

            # exi 5

            return self._get_string_input(mini_spec, nested_list, preset_value, list_idx)

        elif current_attr_type == StringAttributeType.STRING_SET.value:

            # exi 5

            return self._get_string_set_input(mini_spec, nested_list, preset_value, list_idx, self.scenario)

        elif current_attr_type == schema.AttributeType.BLOCK:

            return self.build_scroll_block_edit_widget(spec, nested_list)

    def _get_string_input(self, spec: MiniAttributeSpecs,
                          nested_list, preset_value=None, list_idx=None):
        full_name, attr_type, is_mandatory, _ = spec.get_values()

        line_edit = self._create_line_edit(preset_value, None)

        line_edit.setText(str(preset_value))
        # validator = QtGui.Va()
        # accept '.' as decimal separator
        # validator.setLocale(QtCore.QLocale.Language.English)
        # line_edit.setValidator(validator)

        input_wrapper = InputPanelWrapper(
            inner_widget=line_edit,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
            list_idx=list_idx

        )

        if full_name in self.state_value_dict:
            item = self.state_value_dict[full_name]

            if isinstance(item, list):
                item.append(input_wrapper)
                return line_edit
            new_list = [item]
            new_list.append(input_wrapper)
            self.state_value_dict[full_name] = new_list

            return line_edit

        self.state_value_dict[full_name] = input_wrapper

        return line_edit

    def _get_string_set_input(self, spec: MiniAttributeSpecs,
                              nested_list, preset_value=None, list_idx=None,
                              scenario: object = None):
        full_name, attr_type, is_mandatory, _ = spec.get_values()

        combo_box = QtWidgets.QComboBox()

        sub_string = get_last_substring(full_name)

        string_set_data_dict = get_pre_defined_string_set_values_from_scenario(scenario)

        string_set_data_dict = string_set_data_dict[sub_string]

        partial_result = get_string_set_values_for_group(sub_string)

        string_set_data_dict.extend(partial_result)

        string_set_data_dict = list(set(string_set_data_dict))

        combo_box.addItems(string_set_data_dict)

        input_wrapper = InputPanelWrapper(
            inner_widget=combo_box,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
            list_idx=list_idx

        )

        if full_name in self.state_value_dict:
            item = self.state_value_dict[full_name]

            if isinstance(item, list):
                item.append(input_wrapper)
                return combo_box
            new_list = [item]
            new_list.append(input_wrapper)
            self.state_value_dict[full_name] = new_list

            return combo_box

        self.state_value_dict[full_name] = input_wrapper

        return combo_box

    def _add_list_row(self, spec: AttributeSpecs, widget: QtWidgets.QWidget):
        """Add a new row to a list after the user clicked the add button"""
        widget: CustomListWidget
        block_list = self.build_sub_block_list(spec)

        input_wrapper = InputPanelWrapper(
            inner_widget=block_list,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,

        )

        widget.set_widget(input_wrapper)

        self.ui_parent.updateGeometry()

    def _create_edit_widget(self) -> QtWidgets.QWidget:
        """Create a input panel edit widget for the attribute"""
        spec = self._attr_spec

        if spec.is_list:
            return self.build_list_edit_widget(spec)

        if spec.attr_type == schema.AttributeType.ENUM:

            return self.build_enum_chooser_input_panel(spec)

        elif spec.attr_type == schema.AttributeType.TIME_SERIES:

            return self.build_time_series_edit_widget(spec)

        # INT INPUT PANEL

        elif (spec.attr_type == schema.AttributeType.INTEGER
              or spec.attr_type == schema.AttributeType.LONG):

            return self.build_int_or_long_edit_widget(spec)

        # DOUBLE  INPUT PANEL
        elif spec.attr_type == schema.AttributeType.DOUBLE:

            return self.build_double_edit_widget(spec)

        elif spec.attr_type == schema.AttributeType.BLOCK:

            return self.build_block_edit_widget(spec)

        elif spec.attr_type == schema.AttributeType.STRING:
            return self.build_string_edit_widget(spec)

    def build_block_edit_widget(self, spec: AttributeSpecs):
        block_list = self.build_sub_block_list(spec)

        input_wrapper = InputPanelWrapper(
            inner_widget=block_list,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )
        self.state_value_dict = {spec.full_name: input_wrapper}

        return input_wrapper

    def build_double_edit_widget(self, spec: AttributeSpecs):
        preset_value = self.get_default_value(spec, spec.full_name)

        line_edit = self._create_line_edit(preset_value, None)

        line_edit.setText(str(preset_value))
        validator = QtGui.QDoubleValidator()
        # accept '.' as decimal separator
        validator.setLocale(QtCore.QLocale.English)
        line_edit.setValidator(validator)

        input_wrapper = InputPanelWrapper(
            inner_widget=line_edit,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,

        )

        full_name, attr_type, is_mandatory, _ = spec.get_values()

        if full_name in self.state_value_dict:
            item = self.state_value_dict[full_name]

            if isinstance(item, list):
                item.append(input_wrapper)
                return line_edit
            new_list = [item]
            new_list.append(input_wrapper)
            self.state_value_dict[full_name] = new_list

            return line_edit

        self.state_value_dict[full_name] = input_wrapper

    def build_enum_chooser_input_panel(self, spec: AttributeSpecs):
        combo_box = QtWidgets.QComboBox()
        enum_values = self._attr_enum_values
        combo_box.addItems(enum_values)
        if len(enum_values) == 1:
            combo_box.setCurrentIndex(0)
            self._attr_value = enum_values[0]
        else:
            combo_box.setCurrentIndex(-1)

        preset_value = self.get_default_value(
            spec,
            spec.full_name,
        )

        set_placeholder_text(combo_box, preset_value, enum_values)

        input_wrapper = InputPanelWrapper(
            inner_widget=combo_box,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )

        self.state_value_dict["int_input"] = input_wrapper

        return input_wrapper

    def build_scroll_block_edit_widget(self, spec: AttributeSpecs, nested_list):
        # scroll_area = QtWidgets.QScrollArea()
        scroll_area = QtWidgets.QWidget()
        # scroll_area.setWidgetResizable(True)
        block_list = self.build_sub_block_list(spec, nested_list)

        scroll_area.setLayout(block_list)

        input_wrapper = InputPanelWrapper(
            inner_widget=block_list,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )

        self.state_value_dict[spec.full_name] = input_wrapper
        return input_wrapper

    def build_time_series_edit_widget(self, spec: AttributeSpecs):
        preset_value = self.get_default_value(spec, spec.full_name)

        file_chooser = FileChooserWidgetPanel(
            self._working_dir, spec.attr_type.name, preset_value, self.treeWidget()
        )

        self._display_error = file_chooser.display_error

        input_wrapper = InputPanelWrapper(
            inner_widget=file_chooser,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )

        self.state_value_dict[spec.full_name] = input_wrapper

        return input_wrapper

    def build_int_or_long_edit_widget(self, spec: AttributeSpecs):
        line_edit = self._create_line_edit("0", None)
        preset_value = self.get_default_value(spec, spec.full_name)
        line_edit.setText(str(preset_value))
        line_edit.setValidator(QtGui.QIntValidator())

        input_wrapper = InputPanelWrapper(
            inner_widget=line_edit,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )

        self.state_value_dict[spec.full_name] = input_wrapper
        return input_wrapper

    def build_string_edit_widget(self, spec: AttributeSpecs):
        line_edit = self._create_line_edit("", None)
        preset_value = self.get_default_value(spec, spec.full_name)
        line_edit.setText(preset_value)

        input_wrapper = InputPanelWrapper(
            inner_widget=line_edit,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )

        self.state_value_dict[spec.full_name] = input_wrapper
        return input_wrapper

    def build_list_edit_widget(self, spec: AttributeSpecs):
        preset_value = self.get_nested_list_data()

        if preset_value is None:
            preset_value = []

        for item, idx in zip(preset_value, range(len(preset_value))):
            item: dict
            for k, v in item.items():
                v: Attribute

        if isinstance(preset_value, list):
            dynamic_list = None

            for item, idx in zip(preset_value, range(len(preset_value))):
                if idx == 0:
                    block_list = self.build_sub_block_list(spec, [item])

                    dynamic_list = CustomListWidget(
                        block_list, spec, self.schema, self.attr_name
                    )
                    continue

                item: dict

                block_list = self.build_sub_block_list(spec, [item])

                input_wrapper = InputPanelWrapper(
                    inner_widget=block_list,
                    validation_error_signal=self.validation_error_signal,
                    signal_to_connect=self.validation_signal,
                    attribute=spec,
                )

                dynamic_list.set_widget(input_wrapper)

            if dynamic_list is None:
                block_list = self.build_sub_block_list(spec)

                dynamic_list = CustomListWidget(
                    block_list, spec, self.schema, self.attr_name
                )

            dynamic_list.row_creation_requested.connect(self._add_list_row)

            input_wrapper = InputPanelWrapper(
                inner_widget=dynamic_list,
                validation_error_signal=self.validation_error_signal,
                signal_to_connect=self.validation_signal,
                attribute=spec,
            )

            self.state_value_dict[spec.full_name] = input_wrapper

            return input_wrapper

        block_list = self.build_sub_block_list(spec, preset_value)

        dynamic_list = CustomListWidget(
            block_list, spec, self.schema, self.attr_name
        )

        dynamic_list.row_creation_requested.connect(self._add_list_row)

        input_wrapper = InputPanelWrapper(
            inner_widget=dynamic_list,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )
        self.state_value_dict = {spec.full_name: input_wrapper}

        return input_wrapper

    def search_attr_recursively(self, attr: Attribute, attr_full_name: str):
        """Search for a specific and primitive  nested attribute recursively"""
        for nested_key, nested_item in attr.nested.items():
            nested_item: Attribute

            if attr_full_name == str(nested_key):
                return nested_item

            if nested_item.has_nested:
                temp_result = self.search_attr_recursively(nested_item, attr_full_name)
                if temp_result:
                    return temp_result
                continue

        return None

    def recursive_search(self, data, key):
        """Search for a specific and primitive  nested attribute recursively for any data type"""
        if isinstance(data, list):
            for item in data:
                result = self.recursive_search(item, key)
                if result:
                    return result
        elif isinstance(data, dict):
            if key in data:
                return data[key]
            for k, v in data.items():
                result = self.recursive_search(v, key)
                v: Attribute
                if v.has_nested:
                    result = self.recursive_search(v.nested, key)

                if result:
                    return result
        return None

    def get_nested_list_data(self):
        """Get the nested list data from the root attribute spec
        Since one row can have multiple nested lists,
        we need to get the data from the root attribute spec and then search for the specific nested list data recursively
        """
        if self._default_value:
            if self._default_value.has_nested_list:
                nested_list = self._default_value.nested_list

                return nested_list
        return None

    def is_primitive_attr(self, spec: AttributeSpecs):
        """Check if the Attribute is a final node within the attribute tree /
        Check if the node can be related to a plain value
        """

        if spec.attr_type == schema.AttributeType.BLOCK or spec.attr_type:
            return False
        return True

    def get_default_value(
            self, spec: AttributeSpecs, attr_full_name, nested_list_data=None
    ):
        """Get the saved value for a given attribute spec ( if it exists)"""

        if self.default_agent_flat_value_dict is not None:
            pre = attr_full_name.split('.')[0]
            dict_key = attr_full_name.replace(pre + ".", "")
            if self.is_primitive_attr(spec):
                if dict_key in self.default_agent_flat_value_dict:
                    default_value = self.default_agent_flat_value_dict[dict_key]
                    return default_value

        self._default_value: Attribute

        if nested_list_data:
            matches = re.findall(r"(?<=\.)[^.]*$", attr_full_name)

            attr_short_name = matches[0]

            result = self.recursive_search(nested_list_data, attr_short_name)
            if result:
                return result.value

        if self._default_value:
            if self._default_value.has_nested_list:
                matches = re.findall(r"(?<=\.)[^.]*$", attr_full_name)

                attr_short_name = matches[0]

                nested_list = self._default_value.nested_list

                result = self.recursive_search(nested_list, attr_short_name)
                if result:
                    if result.has_value:
                        return result.value

                return None

        matches = re.findall(r"(?<=\.)[^.]*$", attr_full_name)

        attr_short_name = matches[0]

        if self._default_value:
            if self._default_value.has_nested:
                search_attr_recursively = self.search_attr_recursively(
                    self._default_value, attr_short_name
                )

                if search_attr_recursively:
                    if search_attr_recursively.has_value:
                        return search_attr_recursively.value

        spec_attr_fallback_value = ""

        if not self._default_value:
            return spec_attr_fallback_value

        default_value = ""

        self._default_value: Attribute

        if self._default_value.has_nested:
            for nested_key, nested_item in self._default_value.nested.items():
                nested_item: Attribute

                if attr_full_name.__contains__(str(nested_key)):
                    if nested_item.has_value:
                        default_value = nested_item.value

                    return default_value

        if self._default_value.has_value:
            default_value = self._default_value.value

        default_value = (
            default_value if default_value is not None else spec_attr_fallback_value
        )

        return default_value


    def _create_line_edit(self, placeholder_text: str, handler) -> QtWidgets.QLineEdit:
        """Create a line edit widget for the attribute"""
        line_edit = QtWidgets.QLineEdit()
        line_edit.setText(str(placeholder_text))


        self._display_error = (
            lambda has_error: line_edit.setStyleSheet(
                ERROR_TEXT_STYLE_LINE_EDIT
            )
            if has_error
            else line_edit.setStyleSheet("")
        )
        return line_edit

    def _create_unsupported_edit_widget(self, type_name: str) -> QtWidgets.QLabel:
        """Create a label widget for an unsupported attribute type (Leave for future implementation for new kinds of
        attributes)"""

        logging.error(
            NO_SUPPORTED_EDIT_WIDGET_TEXT.format(
                type_name
            )
        )

        label = QtWidgets.QLabel()

        label.setText(
            self.tr(NO_SUPPORTED_EDIT_WIDGET_TEXT.format(type_name))
        )
        label.setStyleSheet(ERROR_TEXT_STYLE_LABEL)
        return label

    def get_input_value(self):

        state_value_dict = {key: item for key, item in self.state_value_dict.items()}

        if self._attr_spec.attr_type != schema.AttributeType.BLOCK:
            return state_value_dict[next(iter(state_value_dict.keys()))]

        return state_value_dict

    def _get_data_from_input(self):
        """Get the data from the input panel"""
        attr_data = self.get_input_value()

        return attr_data

    def tr(self, msg: str) -> str:
        return QtCore.QCoreApplication.translate("AttributeTreeItem", msg)
